using System;
using System.Linq;
using System.Collections.Generic;

using Model.Cartography.Nav;

namespace Model.Cartography
{
    public class Map
    {
        private NavigationGraph nav = new NavigationGraph();
        public NavigationGraph NavigationGraph { get => nav; }
        public float Scale { get; } /* "world length" of one side of a square in the grid */


        /* grid[0, 0] is the south-western corner, coordinates are as in the first quadrant */
        private Cell[,] grid;
        private int width; /* range on X-axis: [0; width) */
        private int height; /* range on Y-axis: [0; height) */

        private static readonly Orientation[] directions = {
            Orientation.NORTH,
            Orientation.WEST,
            Orientation.SOUTH,
            Orientation.EAST,
        };

        public Map(int width, int height, float scale = 1) {
            if (width <= 0 || height <= 0 || scale <= 0)
                throw new ArgumentOutOfRangeException();

            grid = new Cell[height, width];
            this.width = width;
            this.height = height;
            this.Scale = scale;
        }

        /// <summary>
        /// Builds a <see cref="MapObject" /> using the given <see cref="IMapObjectBluePrint">blueprint</see> at
        /// the given <see cref="MapPosition">position</see>.
        /// </summary>
        /// <param name="blueprint"></param>
        /// <param name="mapPosition"></param>
        /// <returns>the created <see cref="MapObject" /></returns>
        public MapObject Build(IMapObjectBlueprint blueprint, MapPosition mapPosition) {
            if (!CanBuild(blueprint, mapPosition))
                throw new InvalidOperationException();

            var origSize = blueprint.GetSize();
            var size = GetAdjustedSize(origSize, mapPosition.Orientation);

            var obj = blueprint.Place(this, mapPosition.ToMapObjectPosition(Scale));

            for (int dy = 0; dy < size.Y; dy++) {
                for (int dx = 0; dx < size.X; dx++) {
                    int x = mapPosition.Coords.X + dx;
                    int y = mapPosition.Coords.Y + dy;

                    System.Diagnostics.Debug.Assert(grid[y, x].Obj == null);

                    grid[y, x].Obj = obj;
                }
            }

            ConnectNavigableCells(obj, origSize, mapPosition, blueprint.GetNavigableCells());

            return obj;
        }

        private void ConnectNavigableCells(MapObject obj, in Vec2i origSize, in MapPosition mapPosition, IEnumerable<NavigableCell> navcells) {
            if (navcells == null)
                return;

            Orientation rotation = mapPosition.Orientation;

            foreach (var nc in navcells) {
                var p = GetAdjustedMapPosition(origSize, rotation, nc.Position).Add(mapPosition.Coords);

                ref var cell = ref this[p];

                var v = nav.AddVertex(obj);
                obj.Vertices.Add(v);
                cell.Vertex = v;

                if (nc.OpenDirections == null)
                    continue;

                cell.OpenDirections = nc.OpenDirections.Select(d => d.rotate(rotation)).ToList();

                foreach (var dir in cell.OpenDirections) {
                    var neighbor = GetNeighborPos(p, dir);

                    if (!IsInside(neighbor))
                        continue;

                    if (!this[neighbor].Vertex.HasValue)
                        continue;

                    var od = this[neighbor].OpenDirections;

                    if (od == null)
                        continue;

                    // FIXME: inefficient direction checking
                    if (!od.Contains(GetConnectableDirection(dir)))
                        continue;

                    nav.AddEdge(this[p], this[neighbor]);
                }
            }
        }

        private bool IsInside(int x, int y) => 0 <= x && x < width && 0 <= y && y < height;

        private bool IsInside(in Vec2i pos) => IsInside(pos.X, pos.Y);

        private static Vec2i GetNeighborPos(Vec2i pos, Orientation dir) {
            switch (dir) {
                case Orientation.NORTH:
                    pos.Add(0, +1);
                    break;
                case Orientation.WEST:
                    pos.Add(-1, 0);
                    break;
                case Orientation.SOUTH:
                    pos.Add(0, -1);
                    break;
                case Orientation.EAST:
                    pos.Add(+1, 0);
                    break;
            }

            return pos;
        }

        private static Vec2i GetAdjustedMapPosition(in Vec2i size, Orientation rotation, Vec2i pos) {
            switch (rotation) {
                case Orientation.NORTH:
                    /* nothing */
                    break;
                case Orientation.WEST:
                    pos.Rotate90CCW();
                    pos.Add(size.Y - 1, 0);
                    break;
                case Orientation.SOUTH:
                    pos.Rotate180();
                    pos.Add(size.X - 1, size.Y - 1);
                    break;
                case Orientation.EAST:
                    pos.Rotate90CW();
                    pos.Add(0, size.X - 1);
                    break;
            }

            return pos;
        }

        private static Orientation GetConnectableDirection(Orientation dir) {
            /* the eastern neighbor can be connected if it's open to west, etc. */
            return dir.rotate(Orientation.SOUTH); /* 180 deg rotation */
        }

        /* XXX: mapPosition.Coords is the position of the southwestern (bottom left) corner */
        /* XXX: mapPosition.Orientation is the direction the northern side is facing after rotation */
        /*
         * e.g. originally:
         *
         *   north
         *   _____
         *  |  ^  |
         *  |_____| 2
         *     3
         *
         * if mapPosition.Orientation == Orientation.EAST:
         *
         *     __
         *    |  |
         *  3 | >| east
         *    |__|
         *     2
         *
         *  '^' ('>') marks the (originally) northern side of the object
         */
        /* XXX: maybe this should be thought through again... */

        /// <summary>
        /// Checks whether a <see cref="MapObject" /> with the given <see cref="IMapObjectBluePrint">blueprint</see>
        /// can be built at the given <see cref="MapPosition">position</see>.
        /// </summary>
        /// <param name="blueprint"></param>
        /// <param name="mapPosition"></param>
        /// <returns>true if it may be built, false otherwise</returns>
        public bool CanBuild(IMapObjectBlueprint blueprint, MapPosition mapPosition) {
            if (blueprint == null)
                throw new ArgumentOutOfRangeException();

            var size = GetAdjustedSize(blueprint.GetSize(), mapPosition.Orientation);

            if (!IsInside(mapPosition.Coords))
                return false;

            if (!IsInside(mapPosition.Coords.X + size.X - 1, mapPosition.Coords.Y + size.Y - 1))
                return false;

            for (int dy = 0; dy < size.Y; dy++) {
                for (int dx = 0; dx < size.X; dx++) {
                    if (grid[mapPosition.Coords.Y + dy, mapPosition.Coords.X + dx].Obj != null)
                        return false;
                }
            }

            return blueprint.CanPlace(this, mapPosition);
        }

        private static Vec2i GetAdjustedSize(Vec2i size, Orientation rotation) {
            switch (rotation) {
                case Orientation.NORTH:
                case Orientation.SOUTH:
                    /* nothing */
                    break;
                case Orientation.WEST:
                case Orientation.EAST:
                    size.Swap();
                    break;
            }

            return size;
        }

        /// <summary>
        /// Query the <see cref="MapObject" /> that has a given <see cref="VertexHandle">vertex</<see/>.
        /// </summary>
        /// <param name="v"></param>
        /// <returns>the <see cref="MapObject" /></returns>
        public MapObject this[in VertexHandle v] {
            get => nav[v] as MapObject;
        }

        private ref Cell this[in Vec2i pos] {
            get => ref grid[pos.Y, pos.X];
        }

        /// <summary>
        /// Checks whether a certain position on the map has an associated vertex.
        /// </summary>
        /// <param name="pos"></param>
        /// <returns>true if <paramref>pos</paramref> is valid and has a vertex, false otherwise</returns>
        public bool IsNavigable(in Vec2i pos) {
            if (!IsInside(pos))
                return false;

            return this[pos].Vertex.HasValue;
        }

        /// <summary>
        /// Query the MapObject at a given positin.
        /// </summary>
        /// <param name="pos"></param>
        /// <returns>the MapObject</returns>
        public MapObject GetObjectAt(in Vec2i pos) {
            if (!IsInside(pos))
                return null;

            return this[pos].Obj;
        }

        /// <summary>
        /// Queries the valid neighbors of a given position on the map.
        /// </summary>
        /// <param name="pos"></param>
        /// <returns></returns>
        public IEnumerable<Vec2i> GetNeighbors(Vec2i pos) {
            if (!IsInside(pos))
                yield break;

            foreach (var dir in directions) {
                var n = GetNeighborPos(pos, dir);

                if (!IsInside(n))
                    continue;

                yield return n;
            }
        }

        /// <summary>
        /// Maps the canonical blueprint-relative coordinates <paramref>pos</paramref>
        /// to world coordinates.
        /// </summary>
        /// <param name="blueprint"></param>
        /// <param name="mapPosition"></param>
        /// <param name="pos"></param>
        /// <returns></returns>
        public static Vec2i InnerToWorldPosition(IMapObjectBlueprint blueprint, in MapPosition mapPosition, in Vec2i pos) {
            return GetAdjustedMapPosition(blueprint.GetSize(), mapPosition.Orientation, pos).Add(mapPosition.Coords);
        }
    }
}
