using System;
using System.IO;
using System.Linq;
using System.Collections.Generic;
using System.Runtime.CompilerServices;

namespace Model.Cartography.Nav
{
    public class NavigationGraph
    {
        public static readonly int INVALID = -1;

        /// <summary>
        /// Represents the state of a (possibly partial) BFS traversal.
        /// </summary>
        private class TraversalState
        {
            /// <summary>
            /// The generation of the graph in which the traversal was carried out.
            /// </summary>
            public int Gen = INVALID;

            /// <summary>
            /// BFS queue when the traversal was stopped.
            /// </summary>
            public Queue<int> Q = new Queue<int>();
            public int[] Dst, Par; /* Dst is not needed at the moment, but let's keep it for now */

            /// <summary>
            /// Query the next vertex on the path from <paramref>from</paramref>
            /// to the target of the traversal.
            /// </summary>
            /// <param name="from"></param>
            /// <returns></returns>
            public int Next(int from) {
                if (from < 0 || from >= Par.GetLength(0))
                    return INVALID;
                return Par[from];
            }

            public override string ToString() {
                return $"TraversalState {{ Generation = {Gen} }}";
            }
        }

        private class VertexData
        {
            public List<int> Neighbors = new List<int>();

            /* ideally the traversal states would be in some kind of cache and this would some kind of weak reference */
            public TraversalState TraversalFrom = null;

            public object Data = null;

            public override string ToString() {
                return $"VertexData {{ Neighbors = {{ list of {Neighbors.Count} }}, TraversalState = { (TraversalFrom != null ? $"{TraversalFrom}" : "n/a") } }}";
            }
        }

        private int NextID = 0;

        /// <summary>
        /// The generation of the graph, which is used to facilitate reusing of earlier results
        /// if the graph has not changed.
        /// </summary>
        private int Generation = 0;
        private List<VertexData> vertices = new List<VertexData>();

        public int VertexCount { get => NextID; }

#if DEBUG
        public StreamWriter DebugStream { get; set; }
        /* hacks on hacks all the way */

        private void debug(string s, [CallerMemberName] string caller = null)
            => DebugStream?.WriteLine((caller != null ? caller + ": " : "") + s);
#else
        private void debug(string s) { }
#endif

        /// <summary>
        /// Creates a new vertex in the graph. Optionally with arbitrary user data.
        /// </summary>
        /// <param name="data">the arbitrary date to tie to the vertex</param>
        /// <returns>the handle of the newly created vertex</returns>
        public VertexHandle AddVertex(object data = null) {
            debug($"vertex added with id={NextID}, data={data} | Generation = {Generation}");

            if (data != null) {
                EnsureStorage(NextID);
                vertices[NextID].Data = data;
            }

            return new VertexHandle(NextID++);
        }

        /// <summary>
        /// Returns the user data tied to a given vertex in the graph.
        /// </summary>
        /// <param name="v">the vertex</param>
        /// <returns>the user data tied to the vertex</returns>
        public object this[VertexHandle v] {
            get { ValidateVertexHandle(v); if (vertices.Count > v) return vertices[v].Data; else return null; }
            set { ValidateVertexHandle(v); EnsureStorage(v); vertices[v].Data = value; }
        }

        /// <summary>
        /// Creates a undirected edge between two given vertices in the graph.
        /// </summary>
        /// <param name="a">first vertex</param>
        /// <param name="b">second vertex</param>
        public void AddEdge(VertexHandle a, VertexHandle b)
        {
            ValidateVertexHandle(a);
            ValidateVertexHandle(b);

            EnsureStorage(a);
            EnsureStorage(b);

            vertices[a].Neighbors.Add(b);
            vertices[b].Neighbors.Add(a);

            Generation += 1;

            debug($"edge added between {a} and {b} | Generation = {Generation}");
        }

        /// <summary>
        /// Creates and returns a <c>RouteHandle</c> for the path from <c>from</c> to <c>to</c>
        /// </summary>
        /// <param name="from">the first node on the path (starting point)</param>
        /// <param name="to">the last node on the path (destination)</param>
        /// <returns>a <c>RouteHandle</c> for the given route</returns>
        public RouteHandle PlanRoute(VertexHandle from, VertexHandle to) {
            ValidateVertexHandle(from);
            ValidateVertexHandle(to);

            return new RouteHandle(this, from, to);
        }

        /// <summary>
        /// Query the next vertex on the path between two arbitrary vertices in the graph.
        /// </summary>
        /// <param name="from"></param>
        /// <param name="to"></param>
        /// <returns>the handle of the next vertex after <c>from</c> on the path from <c>from</c> to <c>to</c>,
        /// or an invalid vertex if no path is found between the two</returns>
        public VertexHandle GetNextOnRoute(VertexHandle from, VertexHandle to) {
            ValidateVertexHandle(from);
            ValidateVertexHandle(to);

            debug($"get next on route from {from} to {to}");

            EnsureStorage(to);

            /*
             * it is assumed that there are relatively few destinations compared
             * to potential starting vertices, thus routes are calculated *from the destination*
             * towards the starting vertex hoping that partial traversals may be reused
             */
            RecalcIfNecessary(to, from);

            return new VertexHandle(vertices[to].TraversalFrom.Next(from));
        }

        private void RecalcIfNecessary(VertexHandle from, VertexHandle to) {
            ref TraversalState tr = ref vertices[from].TraversalFrom;

            if (tr == null)
                tr = new TraversalState();

            debug($"tr = {tr}");

            if (tr.Gen != Generation) {
                tr.Q.Clear();
                tr.Q.Enqueue(from);
                tr.Dst = Enumerable.Range(0, VertexCount).Select(_ => INVALID).ToArray<int>();
                tr.Dst[from] = 0;
                tr.Par = Enumerable.Range(0, VertexCount).Select(_ => INVALID).ToArray<int>();
                tr.Par[from] = from;

                debug("TraversalState invalidated");
            } else {
                debug("reusing TraversalState");
            }

            if (tr.Next(to) == INVALID) {
                DoTraversal(tr, to);
                debug($"traversal from {from}");
            }
        }

        private void ValidateVertexHandle(VertexHandle v)
        {
            if (v.ID >= NextID || v.ID < 0)
                throw new ArgumentOutOfRangeException();
        }

        private void EnsureStorage(int v)
        {
            if (v >= vertices.Count) {
                vertices.AddRange(Enumerable.Range(0, v - vertices.Count + 1)
                        .Select<int, VertexData>(_ => null));
            }

            if (vertices[v] == null)
                vertices[v] = new VertexData();
        }

        /// <summary>
        /// The core method that does the actual breadth first search.
        /// </summary>
        /// <param name="tr">traversal state to use</param>
        /// <param name="to">target vertex</param>
        private void DoTraversal(TraversalState tr, int to) {
            var q = tr.Q;
            var dst = tr.Dst;
            var par = tr.Par;

            bool found = false;

            while (q.Count > 0 && !found) {
                int u = q.Dequeue();

                debug($"u = {u}");

                foreach (int v in vertices[u].Neighbors) {
                    debug($" v = {v}");
                    if (par[v] != INVALID)
                        continue;
                    debug($" v = {v}");

                    q.Enqueue(v);
                    dst[v] = dst[u] + 1;
                    par[v] = u;

                    if (v == to)
                        found = true;
                }
            }

            tr.Gen = Generation;
        }
    }
}
