using System.Collections.Generic;
using System.Linq;

using Model.Cartography;
using Model.People;

namespace Model.Common
{
    /// <summary>
    /// Maintenance dispacher class, responsible for keeping track of broken buildings and assigning
    /// maintainers the task to fix them.
    /// </summary>
    public static class MaintenanceDispacher
    {
        private static readonly List<MaintenanceStaff> maintainers = new List<MaintenanceStaff>();
        private static readonly Queue<MapObject> backlog = new Queue<MapObject>();

        public static List<MaintenanceStaff> Maintainers => maintainers;

        /// <summary>
        /// Destroys all data.
        /// </summary>
        public static void DestroyInstance() {
            maintainers.Clear();
            backlog.Clear();
        }

        /// <summary>
        /// Adds the broken building to the backlog.
        /// </summary>
        /// <param name="building">the building that broke</param>
        public static void GameBroken(MapObject building) {
            backlog.Enqueue(building);
        }

        /// <summary>
        /// Assigns the task of repairing the broken buildings to the nearest maintainer.
        /// If a building cannot be reached, it will be repaired when possible.
        /// </summary>
        public static void Tick() {
            List<MaintenanceStaff> idle = maintainers.FindAll(staff => staff.Idle).ToList();
            List<MapObject> unreachable = new List<MapObject>();

            while (backlog.Count > 0 && idle.Count > 0) {
                MapObject building = backlog.Dequeue();
                MaintenanceStaff staff = idle.Select((staff, i) => ((staff.Position - building.Position).LengthSquared(), i, staff)).Min().staff;
                staff.Destination = building;
                if (!staff.Idle)
                    idle.Remove(staff);
                else // Staff cannot reach destination
                    unreachable.Add(building);
            }

            unreachable.ForEach(building => backlog.Enqueue(building));
        }
    }
}
