using System.Collections.Generic;

using Model.Cartography;
using Model.Inspection;
using Model.Landscape;
using Model.Utilities;

namespace Model.Common
{
    /// <summary>
    /// Park class, responsible for keeping count of the player's balance and other statistics and
    /// having a list of all the buildings for visitors and others to use.
    /// </summary>
    public class Park : IInspectable, ITickable
    {
        private static Park instance;

        private int balance;
        private Property<int> entryFee = new Property<int>("Entry fee", editable: true);
        private readonly Pavement entrance;
        private readonly List<Pavement> pavements;
        private readonly List<VisitableBuilding> visitableBuildings;
        private readonly List<IStaffBuilding> staffBuildings;

        /// <summary>
        /// Creates instance if not yet created.
        /// </summary>
        public static Park Instance => instance ??= new Park();

        public int Balance => balance;
        public int EntryFee { get => entryFee; set => entryFee.Value = value; }
        public Pavement Entrance => entrance;
        public List<Pavement> Pavements => pavements;
        public List<VisitableBuilding> VisitableBuildings => visitableBuildings;
        public List<IStaffBuilding> StaffBuildings => staffBuildings;

        public string Name => "Park";


        // at the moment, a day is 60 ticks long
        private Property<RollingAggregation<int>> acceptedVisitorsCount =
            new Property<RollingAggregation<int>>("New visitors in the last 3 days", RollingAggregation<int>.CreateRollingSum(60 * 3), false);
        private Property<RollingAggregation<int>> refusedVisitorsCount =
            new Property<RollingAggregation<int>>("Didn't come because of high entry fee", RollingAggregation<int>.CreateRollingSum(60 * 3), false);

        private int acceptedInThisTick = 0;
        private int refusedInThisTick = 0;
        public void CountAcceptedVisitor() => acceptedInThisTick++;
        public void CountRefusedVisitor() => refusedInThisTick++;

        /// <summary>
        /// Constructor of park.
        /// Creates the entrance pavement.
        /// </summary>
        private Park() {
            visitableBuildings = new List<VisitableBuilding>();
            staffBuildings = new List<IStaffBuilding>();
            entrance = (Pavement)Simulation.Instance.Map.Build(Simulation.Settings.EntranceBlueprint, new MapPosition());
            pavements = new List<Pavement> { entrance };
        }

        /// <summary>
        /// Destroys the park instance.
        /// </summary>
        /// <remarks>
        /// Only use this when destroying <see cref="Simulation"/>.
        /// </remarks>
        internal static void DestroyInstance() {
            instance = null;
        }

        /// <summary>
        /// Adds to the player's balance.
        /// </summary>
        /// <param name="amount">the amount to be added</param>
        public void ModifyBalance(int amount) {
            balance += amount;
        }

        /// <summary>
        /// Communicates the tick to every building and the <see cref="MaintenanceDispacher"/>.
        /// Also counts the visitors that were accepted / refused during this period.
        /// </summary>
        public void Tick() {
            visitableBuildings.ForEach(building => building.Tick());
            staffBuildings.ForEach(building => building.Tick());
            MaintenanceDispacher.Tick();

            acceptedVisitorsCount.Value.Roll(acceptedInThisTick);
            acceptedVisitorsCount.Flush();

            refusedVisitorsCount.Value.Roll(refusedInThisTick);
            refusedVisitorsCount.Flush();

            acceptedInThisTick = refusedInThisTick = 0;
        }

        /// <summary>
        /// Gets all the visible properties with <see cref="Inspector.GetProperties(object)"/>.
        /// </summary>
        /// <returns>a list of all the properties</returns>
        public IEnumerable<InspectionElement> GetProperties() => Inspector.GetProperties(this);

        /// <summary>
        /// No commands are available.
        /// </summary>
        /// <returns><c>null</c></returns>
        public IEnumerable<Command> GetCommands() => null;
    }
}
