using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;

using Model.Cartography;
using Model.Inspection;
using Model.Landscape.Blueprints;
using Model.People;

namespace Model.Common
{
    /// <summary>
    /// Simulation class, responsible for providing information between model and view and spawning visitors.
    /// </summary>
    public class Simulation : IBuilder
    {
        private static Simulation instance;

        private readonly Config config;
        private readonly List<Person> people;
        private readonly List<Person> peopleToRemove;
        private readonly Map map;
        private TickResult currentTickResult;
        private int tickCount;

        public static Simulation Instance => instance;
        public static Config Settings => Instance.config;
        public List<Person> People => people;
        public Map Map => map;
        public int TickCount => tickCount;

        public static readonly Random Random = new Random();

        /// <summary>
        /// Constructor of <c>Simulation</c> class.
        /// Ensures the creation of <see cref="Park"/>, sets the default entry fee and initial balance.
        /// </summary>
        /// <param name="config">global settings for the game</param>
        private Simulation(Config config) {
            instance = this;
            this.config = config;

            tickCount = 0;
            people = new List<Person>();
            peopleToRemove = new List<Person>();
            currentTickResult = new TickResult
            {
                newVisitors = new List<(IDisplayable, IInspectable)>(),
                newCleaningStaff = new List<(IDisplayable, IInspectable)>(),
                newMaintenanceStaff = new List<(IDisplayable, IInspectable)>()
            };
            map = new Map(config.GridSize.X, config.GridSize.Y);
            
            Park.Instance.EntryFee = config.DefaultEntryFee; // ensures the creation of Park and entrance
            Park.Instance.ModifyBalance(config.PlayerInitialMoney);
        }

        /// <summary>
        /// Creates the simulation instance.
        /// </summary>
        /// <param name="config">global settings for the game</param>
        /// <returns>the simulation instance created</returns>
        /// <exception cref="Exception">Thrown if simulation is already created.</exception>
        public static Simulation CreateOnce(Config config) {
            if (instance != null)
                throw new Exception("Simulation already created");
            else
                return new Simulation(config);
        }

        /// <summary>
        /// Destroys all the singleton instances.
        /// </summary>
        /// <remarks>
        /// WARNING: Only use this when unloading the level
        /// </remarks>
        public static void DestroyInstance() {
            instance = null;
            Park.DestroyInstance();
            MaintenanceDispacher.DestroyInstance();
        }

        /// <summary>
        /// Returns the balance of the player.
        /// </summary>
        /// <returns>the balance of the player</returns>
        public int GetMoney() {
            return Park.Instance.Balance;
        }

        /// <summary>
        /// Returns the park to be inspected.
        /// </summary>
        /// <returns><see cref="Park.Instance"/> under the <see cref="IInspectable"/> interface.</returns>
        public IInspectable GetPark() {
            return Park.Instance;
        }

        /// <summary>
        /// Returns the entrance of the park.
        /// </summary>
        /// <returns>A tuple containing the entrance twice under two interfaces.</returns>
        public (IDisplayable, IInspectable) GetEntrance() {
            return (Park.Instance.Entrance, Park.Instance.Entrance);
        }

        /// <summary>
        /// Returns <c>this</c>.
        /// </summary>
        /// <returns>Simulation under the <see cref="IBuilder"/> interface.</returns>
        // TODO: maybe remove this
        public IBuilder GetBuilder() {
            return this;
        }

        /// <summary>
        /// Asks the <see cref="Cartography.Map"/> if a building can be built.
        /// </summary>
        /// <param name="blueprint">blueprint of the building to be built</param>
        /// <param name="mapPosition">position of the building to be built</param>
        /// <returns>A boolean value indicating if the building is possible.</returns>
        public bool CanBuild(IMapObjectBlueprint blueprint, MapPosition mapPosition) {
            return map.CanBuild(blueprint, mapPosition);
        }

        /// <summary>
        /// Makes the <see cref="Cartography.Map"/> bulid a building.
        /// </summary>
        /// <param name="blueprint">blueprint of the building to be built</param>
        /// <param name="mapPosition">position of the building to be built</param>
        /// <returns>A tuple containing the <see cref="MapObject"/> twice under two interfaces.</returns>
        public (IDisplayable, IInspectable) Build(IMapObjectBlueprint blueprint, MapPosition mapPosition) {
            var obj = map.Build(blueprint, mapPosition);
            var disp = obj as IDisplayable;
            var insp = obj as IInspectable;
            if (disp == null || insp == null)
                throw new ArgumentException("Building does not implement IDisplayable or IInspectable");
            return (disp, insp);
        }

        /// <summary>
        /// Adds a visitor to the game.
        /// Notifies the view and the park and stores a reference to the person.
        /// Increases balance by entry fee.
        /// </summary>
        /// <param name="person">the visitor to be added</param>
        public void AddPerson(Visitor person) {
            currentTickResult.newVisitors.Add((person, person));
            people.Add(person);
            Park.Instance.ModifyBalance(Park.Instance.EntryFee);
            Park.Instance.CountAcceptedVisitor();
        }

        /// <summary>
        /// Flags a person to be removed at the end of the tick.
        /// </summary>
        /// <param name="person">the person to be removed</param>
        public void RemovePerson(Person person) {
            peopleToRemove.Add(person);
        }

        /// <summary>
        /// Adds a cleaning staff to the game.
        /// Notifies the view and stores a reference to the person.
        /// </summary>
        /// <param name="person">the cleaning staff to be added</param>
        public void AddCleaningStaff(CleaningStaff person) {
            currentTickResult.newCleaningStaff.Add((person, person));
            people.Add(person);
        }

        /// <summary>
        /// Adds a maintenance staff to the game.
        /// Notifies <see cref="MaintenanceDispacher"/> and the view and stores a reference to the person.
        /// </summary>
        /// <param name="person">the maintainer to be added</param>
        public void AddMaintenanceStaff(MaintenanceStaff person) {
            currentTickResult.newMaintenanceStaff.Add((person, person));
            MaintenanceDispacher.Maintainers.Add(person);
            people.Add(person);
        }

        /// <summary>
        /// Counts the ticks, spawns visitors, communicates tick to the park and every person.
        /// Also collects spawns and despawns for the view.
        /// </summary>
        /// <returns>information about spawns and game over</returns>
        public TickResult Tick() {
            tickCount++;

            if (Random.NextDouble() < config.VisitorArrivalChance) {
                int money = Random.Next(config.MinMoney, config.MaxMoney) - Park.Instance.EntryFee;
                if (money >= config.MinMoneyAfterEntryFee)
                    AddPerson(new Visitor(position: config.EntrancePosition,
                                          entryHandle: Park.Instance.Entrance.Vertices.Single(),
                                          speed: config.MovementSpeed,
                                          name: "Example Name",
                                          mood: Random.Next(config.MinMood, config.MaxMood),
                                          satiety: Random.Next(config.MinSatiety, config.MaxSatiety),
                                          money: money));
                else
                    Park.Instance.CountRefusedVisitor();
            }

            Park.Instance.Tick();
            people.ForEach(person => person.Tick());
            peopleToRemove.ForEach(x => people.Remove(x));

            TickResult result = currentTickResult;
            result.gameOver = GetMoney() < 0;

            currentTickResult = new TickResult
            {
                newVisitors = new List<(IDisplayable, IInspectable)>(),
                newCleaningStaff = new List<(IDisplayable, IInspectable)>(),
                newMaintenanceStaff = new List<(IDisplayable, IInspectable)>()
            };
            
            return result;
        }

        /// <summary>
        /// Global settings of the game.
        /// </summary>
        [Serializable]
        public class Config
        {
            public Vector2 EntrancePosition;
            public PavementBlueprint EntranceBlueprint;
            public Vec2i GridSize;
            public int PlayerInitialMoney;
            public float MovementSpeed;
            public float VisitorArrivalChance;
            public int MinMood;
            public int MaxMood;
            public int MinSatiety;
            public int MaxSatiety;
            public int MinMoney;
            public int MaxMoney;
            public int DefaultEntryFee;
            /// <summary>
            /// If any visitors can't spend this amount after the entry fee, they don't come.
            /// </summary>
            public int MinMoneyAfterEntryFee;
            /// <summary>
            /// Maximum amount of garbage removed on each pavement.
            /// </summary>
            public int CleaningStaffCapacity;
        }
    }
}
