using System;
using System.Collections.Generic;

namespace Model.Inspection
{
    /// <summary>
    /// Represents an action that an <see cref="Inspectable" /> can do.
    /// </summary>
    public class Command : InspectionElement
    {
        private Func<object[], object> function;
        private Func<bool> executionCheck;
        private Signature signature;

        public Signature CommandSignature {
            get => signature;
        }

        public bool CanExecute {
            get => executionCheck();
        }

        /// <summary>
        /// Constructs a command with the given name, backing function, signature, and predicate.
        /// </summary>
        /// <param name="name"></param>
        /// <param name="func"></param>
        /// <param name="sign"></param>
        /// <param name="check">the predicate to determine when the function may be called</param>
        public Command(string name, Func<object[],object> func, Signature sign, Func<bool> check) : base(name) {  // TODO: Change it to internal
            executionCheck = check;
            function = func;
            signature = sign;
        }

        /// <summary>
        /// Executes the command with the given arguments.
        /// </summary>
        /// <param name="args"></param>
        /// <returns></returns>
        public object Execute(params object[] args) {
            if (args.Length != signature.Arguments.Count)
                throw new InvalidOperationException("Wrong number of arguments");
            var e1 = signature.Arguments.GetEnumerator();
            var e2 = args.GetEnumerator();
            while (e1.MoveNext()) {
                e2.MoveNext();
                if (!e1.Current.IsAssignableFrom(e2.Current.GetType()))
                    throw new InvalidOperationException($"Wrong argument type. expected: {e1.Current.GetType()}, got: {e2.Current.GetType()}");
            }

            if (CanExecute)
                return function(args);
            else
                throw new InvalidOperationException("Can't execute command");
        }

        /// <summary>
        /// Represents the signature (return type, types of arguments) of a <see cref="Command" />
        /// </summary>
        public struct Signature
        {
            public ICollection<Type> Arguments;
            public Type ReturnType;

            public static Signature Action = new Signature { Arguments = new List<Type>(), ReturnType = typeof(void) };
        }
    }
}
