using System;
using System.Collections.Generic;

namespace Model.Inspection
{
    /// <summary>
    /// Represents a property of an inspectable entity.
    /// </summary>
    /// <typeparam name="T">type of the data this property contains</typeparam>
    public class Property<T> : InspectionElement
    {
        private bool editable;
        public bool IsEditable { get => editable; }

        private Predicate<T> validate;

        private T value;
        public T Value {
            get => value;
            set { // TODO: Add internal access modifier
                this.value = value;
                Changed?.Invoke(this, value);
            }
        }

        public delegate void PropertyChanged(Property<T> sender, T val);
        public event PropertyChanged Changed;

        private List<PropertyChanged> weakEventHandlers = new List<PropertyChanged>();

        /// <summary>
        /// Register a weak event handler to be notified when the value changes.
        /// </summary>
        /// <param name="eventHandler"></param>
        public void RegisterWeakChangedEventHandler(PropertyChanged eventHandler) {
            Changed += eventHandler;
            weakEventHandlers.Add(eventHandler);
        }

        public override void ClearWeakEventHandlers() {
            foreach (var eh in weakEventHandlers)
                Changed -= eh;
            weakEventHandlers.Clear();
        }

        /// <summary>
        /// Changes the value (if editable and the new value is valid).
        /// </summary>
        /// <param name="newValue"></param>
        /// <returns>true if the value was edited successfully, false otherwise</returns>
        public bool Edit(T newValue) {
            if (!editable)
                throw new InvalidOperationException("Property is not editable");

            if (validate?.Invoke(newValue) == false)
                return false;

            Value = newValue;
            return true;
        }

        /// <summary>
        /// Force the view to refresh the displayed value <br/>
        /// (Useful if Value is manipulated outside its setter.)
        /// </summary>
        public void Flush() {
            Changed?.Invoke(this, value);
        }

        internal Property(string name, bool editable = false, Predicate<T> validate = null, bool visible = true) : base(name, visible) {
            this.editable = editable;
            this.validate = validate;
        }

        public Property(string name, T value, bool editable = false, Predicate<T> validate = null, bool visible = true) : this(name, editable, validate, visible) { // TODO: Add internal access modifier
            this.value = value;
        }

        public override string ToString() {
            return value.ToString();
        }

        public static implicit operator T(Property<T> p) => p.Value;
    }
}
