using System;

using Model.Common;
using Model.Cartography;

namespace Model.Landscape.Blueprints
{
    /// <summary>
    /// Blueprint for a <see cref="QueuePavement"/>.
    /// </summary>
    [Serializable]
    public class QueuePavementBlueprint : PavementBlueprint
    {
        public int Capacity;

        /// <summary>
        /// Decides if the <see cref="QueuePavement"/> can be placed.
        /// It can be only placed if it borders exactly one other <see cref="QueuePavement"/> (or it is the first one).
        /// </summary>
        /// <param name="map">same as <see cref="Simulation.Map"/></param>
        /// <param name="mapPosition">position of the pavement</param>
        /// <returns><c>true</c> if the pavement can be placed, <c>false</c> otherwise</returns>
        public override bool CanPlace(Map map, MapPosition mapPosition) {
            if (!base.CanPlace(map, mapPosition))
                return false;

            var d = DetermineDestination(map, mapPosition);
            return d != null;
        }

        /// <summary>
        /// Creates a new <see cref="QueuePavement"/> and adds it to the <see cref="Park.pavements"/> list.
        /// </summary>
        /// <param name="map">same as <see cref="Simulation.Map"/></param>
        /// <param name="position">position of the pavement</param>
        /// <returns>the new <see cref="QueuePavement"/></returns>
        public override MapObject Place(Map map, MapObjectPosition position) {
            var obj = new QueuePavement(this, position, DetermineDestination(map, position));
            Park.Instance.Pavements.Add(obj);
            return obj;
        }

        /// <summary>
        /// Determines the destination of the <see cref="QueuePavement"/> to be created.
        /// </summary>
        /// <param name="map">same as <see cref="Simulation.Map"/></param>
        /// <param name="mapPosition">position of the pavement</param>
        /// <returns>the destination if exactly one possible is found, <c>null</c> otherwise</returns>
        private VisitableBuilding DetermineDestination(Map map, MapPosition mapPosition) {
            VisitableBuilding dst = null;

            // maybe FIXME: this does not take orientation into account
            foreach (var n in map.GetNeighbors(mapPosition.Coords)) {
                switch (map.GetObjectAt(n)) {
                    case QueuePavement queuePavement: /* the following assumes QueuePavements are 1x1 and navigable */
                        if (dst != null && queuePavement.Destination != dst)
                            return null;
                        dst = queuePavement.Destination;
                        break;
                    case IBuilding building:
                        if (map.IsNavigable(n)) {
                            if (dst != null && dst != building)
                                return null;
                            if (!(building is VisitableBuilding vb))
                                return null;
                            dst = vb;
                        }
                        break;
                }
            }

            return dst;
        }
    }
}
