using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;

using Model.Cartography;
using Model.Common;
using Model.Inspection;
using Model.Landscape.Blueprints;
using Model.People;

namespace Model.Landscape
{
    /// <summary>
    /// Interface for unifying generic subtypes of building classes.
    /// </summary>
    public interface IBuilding : ITickable { }

    /// <summary>
    /// Class of buildings, implements construction and interaction with maintenance.
    /// </summary>
    /// <typeparam name="T">type of blueprint for the building</typeparam>
    [Serializable]
    public abstract class Building<T> : MapObject, IBuilding, IDisplayableBuilding where T : BuildingBlueprint
    {
        protected T blueprint;
        protected Property<bool> underConstruction; // protected so that GetFields() will return it (reflection)
        public Property<bool> IsUnderConstruction { get => underConstruction; }
        protected Property<bool> needsRepair;
        public Property<bool> NeedsRepair { get => needsRepair; }
        private int timeUntilOperational;
        private int timeSinceLastRepair;
        private MaintenanceStaff maintainer;
        /// <summary>
        /// The middle of the cell bordering the pavement from where the building can be entered.
        /// </summary>
        protected readonly Vector2 ExitPosition;

        protected Property<int> maintenanceCostRegular;
        protected Property<int> maintenanceInterval;

        public override string Name => blueprint.Name;

        /// <summary>
        /// Building is operational if it is not under construction and is not broken.
        /// </summary>
        public bool IsOperational { get => !underConstruction && !needsRepair; }

        /// <summary>
        /// Constructor of <c>Building</c>.<br/>
        /// Calculates <see cref="ExitPosition"/>.
        /// </summary>
        /// <param name="blueprint">blueprint of building</param>
        /// <param name="position">position of building, see <see cref="MapObject.position"/></param>
        protected Building(T blueprint, MapObjectPosition position) : base(position) {
            this.blueprint = blueprint;
            underConstruction = new Property<bool>("Under construction", value: true);
            needsRepair = new Property<bool>("Broken", value: false);
            timeUntilOperational = blueprint.ConstructionTime;
            timeSinceLastRepair = 0;
            Park.Instance.ModifyBalance(-blueprint.ConstructionCost);

            var entrancePosition = blueprint.GetNavigableCells().Single().Position;
            entrancePosition = Map.InnerToWorldPosition(blueprint, position.MapPosition, entrancePosition);

            ExitPosition = new Vector2(entrancePosition.X + 0.5f, entrancePosition.Y + 0.5f); // XXX: maybe improve

            maintenanceCostRegular = new Property<int>("Regular maintenance cost", blueprint.MaintenanceCost);
            maintenanceInterval = new Property<int>("Regular maintenance interval", blueprint.MaintenanceInterval);
        }

        /// <summary>
        /// Counts down until operational, then deducts maintenance cost.
        /// Decides if building is going to break.
        /// </summary>
        public virtual void Tick() {
            if (timeUntilOperational > 0) {
                timeUntilOperational--;
            } else if (timeUntilOperational == 0) {

                if (Simulation.Instance.TickCount % blueprint.MaintenanceInterval == 0)
                    Park.Instance.ModifyBalance(-blueprint.MaintenanceCost);

                underConstruction.Value = false;
                if (needsRepair) {
                    maintainer.Visible = true;
                    maintainer = null;
                    timeSinceLastRepair = 0;
                }
                needsRepair.Value = false;
                timeSinceLastRepair++;
            }

            // Doesn't break down in the first half of its expected lifespan
            if (!underConstruction.Value
                && !needsRepair
                && timeSinceLastRepair * 2 > blueprint.ExpectedLifeSpan
                && Simulation.Random.Next((blueprint.ExpectedLifeSpan + 1) / 2) == 1)
                Break();
        }

        /// <summary>
        /// Implements logic for breaking.
        /// </summary>
        private void Break() {
            timeUntilOperational = -1;
            needsRepair.Value = true;
            MaintenanceDispacher.GameBroken(this);
        }

        /// <summary>
        /// Implements logic for a person entering the building.
        /// The person will not be visible after entering.
        /// If a maintainer enters while the building is broken, the repair will start
        /// and its cost will be deducted from the player's account.
        /// </summary>
        /// <param name="p">person entering</param>
        /// <param name="pos">position where the person needs to go</param>
        /// <returns>
        /// <see cref="MapObject.EnterResult.SUCCESS"/> if successful,<br/>
        /// <see cref="MapObject.EnterResult.FAIL"/> if not
        /// </returns>
        public override EnterResult Enter(Person p, ref Vector2 pos) {
            p.Visible = false;

            if (p is MaintenanceStaff staff && needsRepair) {
                if(Park.Instance.Balance < blueprint.RepairCost) {
                    MaintenanceDispacher.GameBroken(this);
                    return EnterResult.FAIL;
                }

                Park.Instance.ModifyBalance(-blueprint.RepairCost);

                timeUntilOperational = blueprint.RepairTime;
                maintainer = staff;
            }

            pos = ExitPosition;

            return EnterResult.SUCCESS;
        }

        /// <summary>
        /// Implements logic for a person exiting the building.
        /// The person will be made visible again.
        /// </summary>
        /// <param name="p">the person exiting</param>
        public override void Exit(Person p) {
            p.Visible = true;
            base.Exit(p);
        }

        /// <summary>
        /// Returns an empty list.
        /// </summary>
        /// <returns></returns>
        public override IEnumerable<Command> GetCommands() {
            return new List<Command>();
        }
    }
}
