using System;
using System.Collections.Generic;
using System.Numerics;

using Model.Cartography;
using Model.Common;
using Model.Inspection;
using Model.Landscape.Blueprints;
using Model.People;

namespace Model.Landscape
{
    /// <summary>
    /// Pavement class, directs the randomized movement of people and keeps count of litter and garbage cans.
    /// </summary>
    [Serializable]
    public class Pavement : MapObject
    {
        public readonly Property<bool> HasGarbageCan;
        public readonly Property<int> GarbageAmount;

        private readonly Property<int> garbageCanPrice;

        public override string Name => "Pavement";

        /// <summary>
        /// Constructor of pavements.
        /// Deducts the construction cost from player's account.
        /// </summary>
        /// <remarks>
        /// Only nonzero costs are deducted. This is necessary together with entrance pavement
        /// (<see cref="Simulation.Config.EntranceBlueprint"/>) being free in order to avoid infinite
        /// recursion when constructing <see cref="Park"/>.
        /// </remarks>
        /// <param name="blueprint">blueprint of the pavement</param>
        /// <param name="position">position of the pavement</param>
        public Pavement(PavementBlueprint blueprint, MapObjectPosition position) : base(position) {
            HasGarbageCan   = new Property<bool>("Having garbage can", value: false);
            GarbageAmount   = new Property<int>("Amount of garbage", 0);
            garbageCanPrice = new Property<int>("Price of a garbage can", blueprint.GarbageCanPrice);

            if (blueprint.ConstructionCost > 0)
                Park.Instance.ModifyBalance(-blueprint.ConstructionCost);
        }

        /// <summary>
        /// Directs any person entering to a random position.
        /// </summary>
        /// <param name="p">the person entreing</param>
        /// <param name="pos">the position where the person needs to go</param>
        /// <returns><see cref="MapObject.EnterResult.SUCCESS"/></returns>
        public override EnterResult Enter(Person p, ref Vector2 pos) {
            if (p is Visitor visitor && visitor.ThrowsGarbage() && !HasGarbageCan)
                GarbageAmount.Value++;
            else if (p is CleaningStaff)
                GarbageAmount.Value = Math.Max(0, GarbageAmount - Simulation.Settings.CleaningStaffCapacity);

            pos.X = Position.X + Simulation.Instance.Map.Scale * (float)Simulation.Random.NextDouble();
            pos.Y = Position.Y + Simulation.Instance.Map.Scale * (float)Simulation.Random.NextDouble();

            return EnterResult.SUCCESS;
        }

        /// <summary>
        /// Returns one command, which lets the player add a garbage can to the pavement.
        /// </summary>
        /// <returns>A list of one <see cref="Command"/></returns>
        public override IEnumerable<Command> GetCommands() {
            return new List<Command> {
                new Command(
                    "Add garbage can",
                    (_) => {
                        HasGarbageCan.Value = true;
                        Park.Instance.ModifyBalance(-garbageCanPrice);
                        return null;
                    },
                    Command.Signature.Action,
                    () => !HasGarbageCan && Park.Instance.Balance >= garbageCanPrice)
            };
        }
    }
}
