using System;
using System.Numerics;
using System.Collections.Generic;

using Model.People;
using Model.Inspection;
using Model.Cartography;
using Model.Landscape.Blueprints;

namespace Model.Landscape
{
    /// <summary>
    /// Represents a type of pavement on which a <see cref="Visitor" /> will wait.
    /// </summary>
    [Serializable]
    public class QueuePavement : Pavement
    {
        private QueuePavementBlueprint blueprint;
        public VisitableBuilding Destination { get; }
        public override string Name => "Queue";
        private Property<int> queueingVisitorsCount = new Property<int>("Number of visitors waiting here", 0);
        private List<Visitor> queueingVisitors = new List<Visitor>();

        /*
         * FIXME: two references are stored to the blueprint (one here and one in the base class),
         *        maybe Pavements should be modified similarly to Buildings
         */
        public QueuePavement(QueuePavementBlueprint blueprint, MapObjectPosition position, VisitableBuilding destination)
            : base(blueprint, position)
        {
            this.blueprint = blueprint;
            this.Destination = destination;
        }

        public override EnterResult Enter(Person p, ref Vector2 pos) {
            switch (p) {
                case Visitor visitor:
                    return Enter(visitor, ref pos);
                default:
                    return base.Enter(p, ref pos);
            }
        }

        private EnterResult Enter(Visitor v, ref Vector2 pos) {
            if (v.Destination != this.Destination)
                return base.Enter(v, ref pos);

            if (queueingVisitorsCount < blueprint.Capacity) {
                queueingVisitors.Add(v);
                queueingVisitorsCount.Value = queueingVisitors.Count;

                v.InQueue = true;

                return base.Enter(v, ref pos);
            } else {
                if (v.InQueue)
                    return EnterResult.AGAIN;
                else
                    return EnterResult.FAIL;
            }
        }

        public override void Exit(Person p) {
            switch (p) {
                case Visitor visitor:
                    Exit(visitor);
                    break;
                default:
                    base.Exit(p);
                    break;
            }
        }

        private void Exit(Visitor v) {
            if (queueingVisitors.Remove(v))
                queueingVisitorsCount.Value = queueingVisitors.Count;

            if (v.Destination != this.Destination)
                v.InQueue = false;

            base.Exit(v);
        }
    }
}
