using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;

using Model.Cartography;
using Model.Common;
using Model.Inspection;
using Model.Landscape.Blueprints;
using Model.People;

namespace Model.Landscape
{
    /// <summary>
    /// VisitableBuilding class, handles interaction with visitors.
    /// </summary>
    [Serializable]
    public class VisitableBuilding : Building<VisitableBuildingBlueprint>
    {
        private Property<int> entryFee;
        private Property<int> queueCapacity;
        private Property<int> inQueueCount;
        private Property<int> minVisitorCountToStart;
        private Property<int> rideCapacity;
        private Property<int> onRideCount;
        private Property<int> effectOnMood;
        private Property<int> effectOnSatiety;
        private Property<int> rideDuration;
        private Property<int> remainingRideDuration;
        private Property<int> maintenanceCostPerRide;

        private Queue<Visitor> queue;
        private List<Visitor> onRide;

        public int InQueueCount { get => inQueueCount; } // for testing purposes
        public int OnRideCount { get => onRideCount; } // for testing purposes
        public int EntryFee { get => entryFee; }
        public bool CanEatHere { get => effectOnSatiety > 0; }

        /// <summary>
        /// Constructor of <c>VisitableBuilding</c> class. Sets all the relevant properties
        /// and subscribes to <see cref="entryFee"/>'s <see cref="Property{T}.Changed"/> event.
        /// </summary>
        /// <param name="blueprint"></param>
        /// <param name="position"></param>
        public VisitableBuilding(VisitableBuildingBlueprint blueprint, MapObjectPosition position)
            : base(blueprint, position) {
            entryFee               = new Property<int>("Entry fee", blueprint.EntryFee, true, (x) => x >= 0);
            effectOnMood           = new Property<int>("Effect on mood", blueprint.EffectOnMood);
            effectOnSatiety        = new Property<int>("Effect on satiety", blueprint.EffectOnSatiety);
            minVisitorCountToStart = new Property<int>("Min. number of visitors per ride", blueprint.MinVisitorCountToStart, true, (x) => 0 < x && x <= rideCapacity);
            rideCapacity           = new Property<int>("Max. number of visitors per ride", blueprint.Capacity);
            rideDuration           = new Property<int>("Number of ticks each ride takes", blueprint.TimePerRound);
            queueCapacity          = new Property<int>("Capacity of the queue inside", blueprint.QueueCapacity);
            maintenanceCostPerRide = new Property<int>("Maintenance cost per ride", blueprint.MaintenanceCostPerRide);
            inQueueCount           = new Property<int>("Number of people waiting", 0);
            onRideCount            = new Property<int>("Number of people on the ride", 0);
            remainingRideDuration  = new Property<int>("Remaining ticks from ongoing ride", 0);

            onRide = new List<Visitor>(rideCapacity);
            queue = new Queue<Visitor>(queueCapacity);

            entryFee.Changed += EntryFeeChanged;
        }

        /// <summary>
        /// Removes visitors who cannot pay the new price from the queue.
        /// To be run when the entry fee changes.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="val">the new price</param>
        private void EntryFeeChanged(Property<int> sender, int val) {
            queue = new Queue<Visitor>(queue.Where(visitor => {
                if (visitor.Money < val) {
                    visitor.GameEnded(ExitPosition);
                    return false;
                }
                return true;
            }));

            inQueueCount.Value = queue.Count;
        }

        /// <summary>
        /// Implements the logic for a visitor entering a visitable building.
        /// </summary>
        /// <param name="p">the person entering</param>
        /// <param name="pos"></param>
        /// <returns>
        /// <see cref="MapObject.EnterResult.SUCCESS"/> if successful,<br/>
        /// <see cref="MapObject.EnterResult.FAIL"/> if the queue is full,<br/>
        /// <see cref="MapObject.EnterResult.AGAIN"/> if the visitor is already in an outer queue, but cannot enter yet.
        /// </returns>
        public override EnterResult Enter(Person p, ref Vector2 pos) {
            if (!(p is Visitor))
                return base.Enter(p, ref pos);

            if (!IsOperational)
                return EnterResult.FAIL;

            Visitor visitor = (Visitor)p;

            if (queue.Count == queueCapacity) {
                if (visitor.InQueue)
                    return EnterResult.AGAIN;
                else
                    return EnterResult.FAIL;
            }

            queue.Enqueue(visitor);
            inQueueCount.Value = queue.Count;
            visitor.InQueue = true;

            return base.Enter(p, ref pos);
        }

        /// <summary>
        /// Implements logic for a visitor exiting the building.
        /// </summary>
        /// <param name="p">the visitor exiting</param>
        public override void Exit(Person p) {
            if (p is Visitor visitor && queue.Contains(visitor)) {
                queue = new Queue<Visitor>(queue.Where(v => v != visitor));
                inQueueCount.Value = queue.Count;
                visitor.InQueue = false;
            }

            base.Exit(p);
        }

        /// <summary>
        /// If the building became unoperational, it sends all the visitors away.
        /// If there is an ongoing ride, it counts until that ends.
        /// When it has ended, it modifies the visitors' mood and satiety and lets them exit.
        /// If a new ride can be started (there are enough visitors), they are taken from the queue
        /// and the entry fee is deducted from their accounts.
        /// </summary>
        public override void Tick() {
            base.Tick();

            if (!IsOperational) {
                foreach (Visitor visitor in queue)
                    visitor.GameEnded(ExitPosition);
                queue.Clear();

                foreach(Visitor visitor in onRide) {
                    visitor.ModifyMood(-effectOnMood.Value);
                    visitor.GameEnded(ExitPosition);
                }
                remainingRideDuration.Value = 0;
                onRide.Clear();
                return;
            }

            if (remainingRideDuration > 0) {
                remainingRideDuration.Value--;
            } else {
                onRide.ForEach(visitor => {
                    visitor.ModifyMood(effectOnMood);
                    visitor.ModifySatiety(effectOnSatiety);
                    visitor.GameEnded(ExitPosition);
                });
                onRide.Clear();

                if (queue.Count >= minVisitorCountToStart) {
                    for (int i = 0; i < rideCapacity && queue.Count > 0; i++) {
                        Visitor visitor = queue.Dequeue();
                        visitor.InQueue = false;
                        visitor.ModifyMoney(-entryFee);
                        onRide.Add(visitor);
                    }

                    Park.Instance.ModifyBalance(onRide.Count * EntryFee - maintenanceCostPerRide);
                    remainingRideDuration.Value = rideDuration;
                }
            }

            onRideCount.Value = onRide.Count;
            inQueueCount.Value = queue.Count;
        }
    }
}
