using System.Linq;
using System.Numerics;

using Model.Cartography;
using Model.Cartography.Nav;
using Model.Common;
using Model.Inspection;

namespace Model.People
{
    /// <summary>
    /// Person class, responsible for the movement of people.
    /// </summary>
    public class Person : WorldEntity, IInspectable, ITickable
    {
        private MapObject destination;
        private Vector2 nextPosition;
        private readonly float speed;
        protected readonly Property<string> name;
        private RouteHandle routeHandle;
        private VertexHandle nextVertexHandle = VertexHandle.INVALID;

        /// <summary>
        /// The destination of the person. Setting this to a new value causes
        /// the <see cref="routeHandle"/> to be updated and the next step (vertex) to be calculated.
        /// </summary>
        public virtual MapObject Destination {
            get => destination;
            set {
                if (destination == value) // avoid recalculating (and stepping in navgraph)
                    return;

                destination = value;
                if (value == null)
                    return;

                routeHandle = Simulation.Instance.Map.NavigationGraph.PlanRoute(nextVertexHandle, value.Vertices.Single());

                CalculateNextStep();
            }
        }

        /// <summary>
        /// Gets the <see cref="Cartography.MapObject"/> the Person is at.
        /// </summary>
        /// <remarks>
        /// Makes the <see cref="Map"/> convert the <see cref="nextVertexHandle"/> to <see cref="Cartography.MapObject"/>.
        /// </remarks>
        protected MapObject MapObject { get => Simulation.Instance.Map[nextVertexHandle]; }

        /// <summary>
        /// Constructor of Person class.
        /// </summary>
        /// <param name="position">starting world position</param>
        /// <param name="entryHandle">starting <c>VertexHandle</c></param>
        /// <param name="speed">speed (max. distance travelled on each tick)</param>
        /// <param name="name">name</param>
        protected Person(Vector2 position, VertexHandle entryHandle, float speed, string name = null) : base(position) {
            this.speed = speed;
            this.name = new Property<string>("Name", name);
            nextVertexHandle = entryHandle;
            Visible = true;
        }

        /// <summary>
        /// Steps the person if visible. To be called on each tick.
        /// </summary>
        public virtual void Tick() {
            if (Visible)
                StepPerson();
        }

        private void StepPerson() {
            if (destination == null)
                return;

            Vector2 toGoal = nextPosition - Position;

            if (toGoal.Length() <= speed) {
                Position = nextPosition;
                CalculateNextStep();
            } else {
                Position += toGoal / toGoal.Length() * speed;
            }
        }

        /// <summary>
        /// Removes the destination. To be called on arrival.
        /// </summary>
        /// <param name="dest">destination</param>
        protected virtual void Arrived(MapObject dest) {
            destination = null;
        }

        private void CalculateNextStep() {
            if (routeHandle.Arrived) {
                Arrived(destination);
                return;
            }

            var nextPossibleVert = routeHandle.GetNext();

            if (!nextPossibleVert.Valid) {
                destination = null;
                return;
            }

            MapObject nextObj = Simulation.Instance.Map[nextPossibleVert];
            Vector2 pos = new Vector2();

            switch (nextObj.Enter(this, ref pos)) {
                case MapObject.EnterResult.FAIL:
                    destination = null;
                    return;
                case MapObject.EnterResult.AGAIN:
                    return;
                case MapObject.EnterResult.SUCCESS:
                    /* nothing */
                    break;
            }

            MapObject.Exit(this);

            nextVertexHandle = routeHandle.CommitStep();
            nextPosition = pos;
        }

        /// <summary>
        /// Makes the person exit the park, removing it from the list of people and deleting its handle.
        /// </summary>
        protected void ExitPark() {
            Simulation.Instance.RemovePerson(this);
            Handle.Delete(null);
        }
    }
}
