using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;

using Model.Cartography;
using Model.Cartography.Nav;
using Model.Common;
using Model.Inspection;
using Model.Landscape;

namespace Model.People
{
    /// <summary>
    /// Visitor class, implements the logic for visitors arriving and being in the park.
    /// </summary>
    public class Visitor : Person
    {
        private Property<int> money;
        private Property<int> mood;
        private Property<int> satiety;
        private Property<bool> inQueue;
        private int stepsSinceLastMeal;

        public bool InQueue { get => inQueue.Value; set => inQueue.Value = value; }
        public int Mood { get => mood; } // For testing purposes
        public int Satiety { get => satiety; } // For testing purposes
        public int Money { get => money; }
        public override string Name => "Visitor";

        /// <summary>
        /// Constructor of visitors
        /// </summary>
        /// <param name="position"></param>
        /// <param name="entryHandle"></param>
        /// <param name="speed"></param>
        /// <param name="name"></param>
        /// <param name="mood">starting mood</param>
        /// <param name="satiety">starting satiety</param>
        /// <param name="money">starting money</param>
        public Visitor(Vector2 position, VertexHandle entryHandle, float speed, string name, int mood, int satiety, int money) : base(position, entryHandle, speed, name) {
            this.mood = new Property<int>("Mood", mood);
            this.satiety = new Property<int>("Satiety", satiety);
            this.money = new Property<int>("Money", money);
            inQueue = new Property<bool>("In queue", value: false);
            stepsSinceLastMeal = -1;
            ChooseDestination();
        }

        /// <summary>
        /// Chooses an operational building as destination, perferring lower prices
        /// and places where they can eat if their hunger is low.
        /// </summary>
        private void ChooseDestination() {
            if (mood < 0) {
                Destination = Park.Instance.Entrance;
                return;
            }

            List<VisitableBuilding> possibilities = Park.Instance.VisitableBuildings
                .Where(building => building.IsOperational).ToList();

            if (satiety < 0)
                possibilities = possibilities.FindAll(building => building.CanEatHere);

            if (possibilities.Count == 0)
                return; // Return early as there is nothing to choose from

            List<int> weights = possibilities.Select(building => money / Math.Max(1, building.EntryFee)).ToList();

            int sum = weights.Sum();

            if (sum <= 0) {
                Destination = Park.Instance.Entrance;
                return;
            }

            int choice = Simulation.Random.Next(sum);

            Destination = possibilities[weights.FindIndex(value => { choice -= value; return choice < 0; })];
        }

        /// <summary>
        /// Signals the end of the game.
        /// Makes the visitor exit to the street and look for a new destination.
        /// </summary>
        /// <param name="exit">world position of the exit</param>
        public void GameEnded(Vector2 exit) {
            inQueue.Value = false;
            Position = exit;
            Visible = true;
            ChooseDestination();
        }

        /// <summary>
        /// Makes the person decide if it throws garbage away.
        /// This is more likely if the person has recently eaten.
        /// To be called when entering a pavement.
        /// </summary>
        /// <returns><c>true</c> if the person throws a piece of garbage away</returns>
        public bool ThrowsGarbage() {
            if (stepsSinceLastMeal == -1)
                return false;

            return Simulation.Random.Next(++stepsSinceLastMeal + 2) == 0;
        }

        /// <summary>
        /// Updates the visitor's mood and satiety, steps the person, and makes it
        /// choose a new destination if necessary.
        /// </summary>
        public override void Tick() {
            base.Tick();

            if (Destination == null)
                ChooseDestination();

            int moodChange = -1;

            if (inQueue)
                moodChange *= 2;
            if (satiety < 0)
                moodChange *= 2;
            if (MapObject is Pavement pavement)
                moodChange -= Simulation.Random.Next(pavement.GarbageAmount / 2);

            mood.Value += moodChange;

            satiety.Value -= 1;

            if (mood < 0 && (Visible || inQueue))
                Destination = Park.Instance.Entrance;
        }

        /// <summary>
        /// Adds to the visitor's mood.
        /// </summary>
        /// <param name="value">the amount to be added</param>
        public void ModifyMood(int value) {
            mood.Value += value;
        }

        /// <summary>
        /// Adds to the visitor's satiety.
        /// </summary>
        /// <param name="value">the amount to be added</param>
        public void ModifySatiety(int value) {
            satiety.Value += value;

            if (value > 0)
                stepsSinceLastMeal = 0;
        }

        /// <summary>
        /// Adds to the visitor's money.
        /// </summary>
        /// <param name="value">the amount to be added</param>
        public void ModifyMoney(int value) {
            money.Value += value;
        }

        /// <summary>
        /// Makes the visitor leave the park if the exit is reached.
        /// </summary>
        /// <param name="dest">the destination reached</param>
        protected override void Arrived(MapObject dest) {
            if (dest == Park.Instance.Entrance && MapObject == dest)
                ExitPark();
            base.Arrived(dest);
        }
    }
}
