using System.Collections.Generic;

namespace Model.Utilities
{
    /// <summary>
    /// Class implementing rolling aggregation, remembering and adding up the last few data points
    /// to provide statistical information over time.
    /// </summary>
    /// <typeparam name="T">the type of data to aggregate</typeparam>
    public class RollingAggregation<T> 
    {
        private readonly int windowSize;
        private readonly Queue<T> queue;

        public delegate void Addition(ref T group, T item);
        public delegate void Substraction(ref T group, T item);

        private readonly Addition add;
        private readonly Substraction substract;

        private T sum;

        /// <summary>
        /// The current sum of items in the window.
        /// </summary>
        public T Sum => sum;

        /// <summary>
        /// The current number of items in the window.
        /// </summary>
        public int Count => queue.Count;

        /// <summary>
        /// Constructor of <see cref="RollingAggregation{T}"/>.
        /// </summary>
        /// <param name="windowSize">number of previous values to be remembered</param>
        /// <param name="add">operation to use on a new element to be added</param>
        /// <param name="substract">operation to use on an old element to be removed</param>
        public RollingAggregation(int windowSize, Addition add, Substraction substract) {
            queue = new Queue<T>(windowSize+1);
            this.windowSize = windowSize;
            this.add = add;
            this.substract = substract;
            sum = default(T);
        }

        /// <summary>
        /// Rolls the "sliding window" by adding val.
        /// </summary>
        /// <param name="val">the new value to enter the window</param>
        /// <returns>the current sum of the window</returns>
        public T Roll(T val) {
            add(ref sum, val);
            queue.Enqueue(val);
            if (Count > windowSize)
                substract(ref sum, queue.Dequeue());
            return sum;
        }

        /// <summary>
        /// Clears the previously remembered values.
        /// </summary>
        public void Clear() {
            sum = default(T);
            queue.Clear();
        }

        /// <summary>
        /// Returns the current sum as a string.
        /// </summary>
        /// <returns>the current sum as a string</returns>
        public override string ToString() {
            return sum.ToString();
        }

        /// <summary>
        /// Retrives the sum of the <see cref="RollingAggregation{T}"/> provided.
        /// </summary>
        /// <param name="ra">the <see cref="RollingAggregation{T}"/> storing the sum</param>
        public static implicit operator T(RollingAggregation<T> ra) {
            return ra.sum;
        }


        /// <summary>
        /// Creates a rolling sum.
        /// </summary>
        /// <remarks>
        /// Provided for ease of use.
        /// </remarks>
        /// <param name="windowSize">number of previous values to be remembered</param>
        /// <returns>a new <see cref="RollingAggregation{T}"/></returns>
        public static RollingAggregation<int> CreateRollingSum(int windowSize) {
            return new RollingAggregation<int>(
                windowSize,
                (ref int g, int i) => g += i,
                (ref int g, int i) => g -= i
            );
        }
    }
}
