using System;
using System.Linq;
using System.Collections.Generic;

using NUnit.Framework;

using Model.Cartography;

namespace Tests.Model.Cartography {
    public class Test_Map
    {
        private class TestMapObjectBlueprint : IMapObjectBlueprint
        {
            public Func<Vec2i> GetSizeFunc;
            public Func<IEnumerable<NavigableCell>> GetNavigableCellsFunc;
            public Func<Map, MapPosition, bool> CanPlaceFunc;
            public Func<Map, MapObjectPosition, MapObject> PlaceFunc;

            public Vec2i GetSize() => GetSizeFunc();
            public IEnumerable<NavigableCell> GetNavigableCells() => GetNavigableCellsFunc?.Invoke();
            public bool CanPlace(Map map, MapPosition mapPosition) => CanPlaceFunc?.Invoke(map, mapPosition) ?? true;
            public MapObject Place(Map map, MapObjectPosition mapObjectPosition) => PlaceFunc(map, mapObjectPosition);
        }

        private class DummyMapObject : MapObject
        {
            public DummyMapObject(MapObjectPosition mapObjectPosition) : base(mapObjectPosition) { }
        }

        [Test]
        public void TestCanPlace_False()
        {
            Map m = new Map(10, 10);

            Assert.IsFalse(m.CanBuild(
                new TestMapObjectBlueprint() {
                    GetSizeFunc = () => new Vec2i(1, 1),
                    CanPlaceFunc = (m, p) => false,
                },
                new MapPosition() {
                    Coords = new Vec2i(0, 0),
                    Orientation = Orientation.NORTH,
                }
            ));
        }

        [Test]
        public void TestCanPlace_TooBig() {
            Map m = new Map(10, 10);

            Assert.IsFalse(m.CanBuild(
                new TestMapObjectBlueprint() {
                    GetSizeFunc = () => new Vec2i(20, 20),
                },
                new MapPosition() {
                    Coords = new Vec2i(0, 0),
                    Orientation = Orientation.NORTH,
                }
            ));

            Assert.IsFalse(m.CanBuild(
                new TestMapObjectBlueprint() {
                    GetSizeFunc = () => new Vec2i(6, 6),
                },
                new MapPosition() {
                    Coords = new Vec2i(5, 5),
                    Orientation = Orientation.NORTH,
                }
            ));
        }

        [Test]
        public void TestCanPlace_TopOfOther() {
            Map m = new Map(10, 10);

            Assert.NotNull(m.Build(
                new TestMapObjectBlueprint() {
                    GetSizeFunc = () => new Vec2i(1, 1),
                    PlaceFunc = (m, p) => new DummyMapObject(p),
                },
                new MapPosition() {
                    Coords = new Vec2i(5, 5),
                    Orientation = Orientation.NORTH,
                }
            ));

            Assert.IsFalse(m.CanBuild(
                new TestMapObjectBlueprint() {
                    GetSizeFunc = () => new Vec2i(1, 1),
                },
                new MapPosition() {
                    Coords = new Vec2i(5, 5),
                    Orientation = Orientation.NORTH,
                }
            ));
        }

        [Test]
        public void TestCanPlace_Intersecting() {
            Map m = new Map(10, 10);

            Assert.NotNull(m.Build(
                new TestMapObjectBlueprint() {
                    GetSizeFunc = () => new Vec2i(2, 3),
                    PlaceFunc = (m, p) => new DummyMapObject(p),
                },
                new MapPosition() {
                    Coords = new Vec2i(5, 5),
                    Orientation = Orientation.NORTH,
                }
            ));

            Assert.IsFalse(m.CanBuild(
                new TestMapObjectBlueprint() {
                    GetSizeFunc = () => new Vec2i(2, 2),
                },
                new MapPosition() {
                    Coords = new Vec2i(4, 4),
                    Orientation = Orientation.NORTH,
                }
            ));
        }

        [Test]
        public void TestCanPlace_NonIntersecting() {
            Map m = new Map(10, 10);

            Assert.NotNull(m.Build(
                new TestMapObjectBlueprint() {
                    GetSizeFunc = () => new Vec2i(2, 3),
                    PlaceFunc = (m, p) => new DummyMapObject(p),
                },
                new MapPosition() {
                    Coords = new Vec2i(5, 5),
                    Orientation = Orientation.NORTH,
                }
            ));

            Assert.IsTrue(m.CanBuild(
                new TestMapObjectBlueprint() {
                    GetSizeFunc = () => new Vec2i(2, 2),
                },
                new MapPosition() {
                    Coords = new Vec2i(3, 3),
                    Orientation = Orientation.NORTH,
                }
            ));
        }

        [Test]
        public void TestBuild() {
            Map m = new Map(10, 10);

            var b = m.Build(
                new TestMapObjectBlueprint() {
                    GetSizeFunc = () => new Vec2i(2, 3),
                    PlaceFunc = (m, p) => new DummyMapObject(p),
                },
                new MapPosition() {
                    Coords = new Vec2i(5, 5),
                    Orientation = Orientation.NORTH,
                }
            );

            Assert.NotNull(b);

            Assert.AreEqual(b, m.GetObjectAt(new Vec2i(5, 5)));
        }

        [Test]
        public void TestNavigableCells() {
            Map m = new Map(10, 10);

            // -------------------------------

            var b1 = m.Build(
                new TestMapObjectBlueprint() {
                    GetSizeFunc = () => new Vec2i(1, 1),
                    GetNavigableCellsFunc = () => new List<NavigableCell>() {
                        new NavigableCell() {
                            Position = new Vec2i(0, 0),
                            OpenDirections = new List<Orientation>() { Orientation.NORTH, Orientation.SOUTH },
                        },
                    },
                    PlaceFunc = (m, p) => new DummyMapObject(p),
                },
                new MapPosition() {
                    Coords = new Vec2i(5, 5),
                    Orientation = Orientation.EAST,
                }
            );

            Assert.NotNull(b1);
            Assert.NotNull(b1.Vertices);
            Assert.AreEqual(b1.Vertices.Count, 1);
            Assert.AreEqual(b1, m[b1.Vertices.Single()]);

            // -------------------------------

            var b2 = m.Build(
                new TestMapObjectBlueprint() {
                    GetSizeFunc = () => new Vec2i(2, 1),
                    GetNavigableCellsFunc = () => new List<NavigableCell>() {
                        new NavigableCell() {
                            Position = new Vec2i(1, 0),
                            OpenDirections = new List<Orientation>() { Orientation.EAST },
                        },
                    },
                    PlaceFunc = (m, p) => new DummyMapObject(p),
                },
                new MapPosition() {
                    Coords = new Vec2i(3, 5),
                    Orientation = Orientation.NORTH,
                }
            );

            Assert.NotNull(b2);
            Assert.NotNull(b2.Vertices);
            Assert.AreEqual(b2.Vertices.Count, 1);
            Assert.AreEqual(b2, m[b2.Vertices.Single()]);

            // -------------------------------

            var b3 = m.Build(
                new TestMapObjectBlueprint() {
                    GetSizeFunc = () => new Vec2i(1, 1),
                    GetNavigableCellsFunc = () => new List<NavigableCell>() {
                        new NavigableCell() {
                            Position = new Vec2i(0, 0),
                            OpenDirections = new List<Orientation>() { Orientation.EAST },
                        },
                    },
                    PlaceFunc = (m, p) => new DummyMapObject(p),
                },
                new MapPosition() {
                    Coords = new Vec2i(6, 5),
                    Orientation = Orientation.NORTH,
                }
            );

            Assert.NotNull(b3);
            Assert.NotNull(b3.Vertices);
            Assert.AreEqual(b3.Vertices.Count, 1);
            Assert.AreEqual(b3, m[b3.Vertices.Single()]);

            // -------------------------------

            Assert.IsTrue(m.NavigationGraph.GetNextOnRoute(b1.Vertices.Single(), b2.Vertices.Single()).Valid);
            Assert.IsFalse(m.NavigationGraph.GetNextOnRoute(b1.Vertices.Single(), b3.Vertices.Single()).Valid);
            Assert.IsFalse(m.NavigationGraph.GetNextOnRoute(b2.Vertices.Single(), b3.Vertices.Single()).Valid);
        }
    }
}
