using NUnit.Framework;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;

using Model.Cartography;
using Model.Common;
using Model.Landscape;
using Model.Landscape.Blueprints;
using Model.People;

using static Model.Cartography.MapObject;

namespace Tests.Model.Landscape
{
    class Test_VisitableBuilding
    {
        private Simulation.Config config = new Simulation.Config
        {
            VisitorArrivalChance = 0,
            EntranceBlueprint = new PavementBlueprint
            {
                Size = new Vec2i { X = 1, Y = 1 },
                NavigableCells = new List<NavigableCell> {
                    new NavigableCell {
                        Position = new Vec2i { X = 0, Y = 0 },
                        OpenDirections = new List<Orientation> {
                            Orientation.NORTH,
                            Orientation.EAST,
                            Orientation.SOUTH,
                            Orientation.WEST,
                        },
                    },
                },
            },
            EntrancePosition = new Vector2(0, 0),
            GridSize = new Vec2i { X = 1, Y = 1 },
        };

        VisitableBuildingBlueprint blueprint = new VisitableBuildingBlueprint
        {
            EntryFee = 1,
            QueueCapacity = 1,
            Capacity = 1,
            EffectOnMood = 2,
            EffectOnSatiety = 1,
            NavigableCells = new List<NavigableCell> {
                new NavigableCell {
                    Position = new Vec2i()
                }
            }
        };

        VisitableBuilding building;

        Vector2 vec = new Vector2();
        Visitor visitor;

        [SetUp]
        public void Init() {
            Simulation.CreateOnce(config);

            building = new VisitableBuilding(blueprint, new MapObjectPosition());

            building.Tick();

            Assert.IsTrue(building.IsOperational);

            visitor = new Visitor(vec, Park.Instance.Entrance.Vertices.Single(), 0, "", 1, 1, 1);
        }

        [TearDown]
        public void TearDown() {
            Simulation.DestroyInstance();
        }

        [Test]
        public void TestCanEatHere() {
            Assert.AreEqual(building.CanEatHere, blueprint.EffectOnSatiety > 0);
        }

        [Test]
        public void TestEntering() {
            Assert.IsTrue(visitor.Visible);
            Assert.IsFalse(visitor.InQueue);
            Assert.AreEqual(0, building.InQueueCount);

            Assert.AreEqual(EnterResult.SUCCESS, building.Enter(visitor, ref vec));

            Assert.IsFalse(visitor.Visible);
            Assert.IsTrue(visitor.InQueue);
            Assert.AreEqual(1, building.InQueueCount);
        }

        [Test]
        public void TestRide() {
            Assert.AreEqual(EnterResult.SUCCESS, building.Enter(visitor, ref vec));

            Assert.IsTrue(visitor.InQueue);
            Assert.AreEqual(1, building.InQueueCount);
            Assert.AreEqual(0, building.OnRideCount);

            building.Tick();

            Assert.IsFalse(visitor.InQueue);
            Assert.AreEqual(0, building.InQueueCount);
            Assert.AreEqual(1, building.OnRideCount);
        }

        [Test]
        public void TestEffects() {
            int mood = visitor.Mood;
            int satiety = visitor.Satiety;
            int money = visitor.Money;

            Assert.AreEqual(EnterResult.SUCCESS, building.Enter(visitor, ref vec));

            building.Tick();
            building.Tick();

            Assert.IsTrue(visitor.Visible);
            Assert.AreEqual(mood + blueprint.EffectOnMood, visitor.Mood);
            Assert.AreEqual(satiety + blueprint.EffectOnSatiety, visitor.Satiety);
            Assert.AreEqual(money - blueprint.EntryFee, visitor.Money);
        }

        [Test]
        public void TestExitingFromQueue() {
            Assert.AreEqual(EnterResult.SUCCESS, building.Enter(visitor, ref vec));

            Assert.IsFalse(visitor.Visible);
            Assert.IsTrue(visitor.InQueue);
            Assert.AreEqual(1, building.InQueueCount);

            building.Exit(visitor);

            Assert.IsTrue(visitor.Visible);
            Assert.IsFalse(visitor.InQueue);
            Assert.AreEqual(0, building.InQueueCount);
        }
    }
}
