using NUnit.Framework;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;

using Model.Cartography;
using Model.Common;
using Model.Landscape;
using Model.Landscape.Blueprints;
using Model.People;


namespace Tests.Model.People
{
    class Test_Visitor
    {
        private Simulation.Config config = new Simulation.Config
        {
            VisitorArrivalChance = 1,
            MinMoney = 1,
            MaxMoney = 2,
            MinMoneyAfterEntryFee = 1,
            MinMood = 3,
            MaxMood = 5,
            MinSatiety = 2,
            MaxSatiety = 3,
            EntranceBlueprint = new PavementBlueprint
            {
                Size = new Vec2i { X = 1, Y = 1 },
                NavigableCells = new List<NavigableCell> {
                    new NavigableCell {
                        Position = new Vec2i { X = 0, Y = 0 },
                        OpenDirections = new List<Orientation> {
                            Orientation.NORTH,
                            Orientation.EAST,
                            Orientation.SOUTH,
                            Orientation.WEST,
                        },
                    },
                },
            },
            EntrancePosition = new Vector2(0, 0),
            GridSize = new Vec2i { X = 2, Y = 1 },
        };

        private Visitor visitor;

        [SetUp]
        public void Init() {
            if (Simulation.Instance == null) {
                Simulation.CreateOnce(config);
            } else {
                Simulation.Instance.People.Clear();
            }

            Simulation.Instance.Tick();

            Assert.AreEqual(1, Simulation.Instance.People.Count);

            visitor = (Visitor)Simulation.Instance.People.Single();

            Park.Instance.EntryFee = 0;
        }

        [TearDown]
        public void TearDown() {
            Simulation.DestroyInstance();
        }

        [Test]
        public void TestSpawning() {
            visitor.ModifyMood(1); // compensate for one tick
            visitor.ModifySatiety(1);

            Assert.GreaterOrEqual(config.MaxMoney, visitor.Money);
            Assert.LessOrEqual(config.MinMoney, visitor.Money);
            Assert.GreaterOrEqual(config.MaxMood, visitor.Mood);
            Assert.LessOrEqual(config.MinMood, visitor.Mood);
            Assert.GreaterOrEqual(config.MaxSatiety, visitor.Satiety);
            Assert.LessOrEqual(config.MinSatiety, visitor.Satiety);
        }

        [Test]
        public void TestLeaving() {
            visitor.ModifySatiety(10); // should be more, than enough
            visitor.ModifyMood(1 - visitor.Mood);

            Assert.AreEqual(1, visitor.Mood);
            if (Park.Instance.VisitableBuildings.Count == 0)
                Assert.IsNull(visitor.Destination);

            Simulation.Instance.Tick();

            Assert.AreEqual(0, visitor.Mood);
            if (Park.Instance.VisitableBuildings.Count == 0)
                Assert.IsNull(visitor.Destination);

            Simulation.Instance.Tick();

            Assert.AreEqual(-1, visitor.Mood);
            // FIXME: Destination is the current position, so it cannot stay destination
            //Assert.AreEqual(Park.Instance.Entrance, visitor.Destination);
        }

        [Test]
        public void TestHunger() {
            visitor.ModifySatiety(1 - visitor.Satiety);
            int mood = visitor.Mood;

            Assert.AreEqual(1, visitor.Satiety);

            Simulation.Instance.Tick();

            Assert.AreEqual(mood - 1, visitor.Mood);
            Assert.AreEqual(0, visitor.Satiety);

            Simulation.Instance.Tick();

            Assert.AreEqual(mood - 2, visitor.Mood);
            Assert.AreEqual(-1, visitor.Satiety);

            Simulation.Instance.Tick();

            Assert.AreEqual(mood - 4, visitor.Mood);
        }

        [Test]
        public void TestQueue() {
            int mood = visitor.Mood;

            Simulation.Instance.Tick();

            Assert.AreEqual(mood - 1, visitor.Mood);

            visitor.InQueue = true;

            Simulation.Instance.Tick();

            Assert.AreEqual(mood - 3, visitor.Mood);
        }

        [Test]
        public void TestDestination() {
            if (Park.Instance.VisitableBuildings.Count > 0) {
                Simulation.Instance.Tick();
                Assert.IsNotNull(visitor.Destination);
                return;
            }

            Assert.IsNull(visitor.Destination);

            VisitableBuilding testBuilding = (VisitableBuilding)Simulation.Instance.Map.Build(
                new VisitableBuildingBlueprint
                {
                    Size = new Vec2i { X = 1, Y = 1 },
                    NavigableCells = new List<NavigableCell> {
                        new NavigableCell {
                            Position = new Vec2i { X = 0, Y = 0 },
                            OpenDirections = new List<Orientation> {
                                Orientation.NORTH,
                                Orientation.EAST,
                                Orientation.SOUTH,
                                Orientation.WEST
                            },
                        },
                    },
                    QueueCapacity = 1,
                },
                new MapPosition
                {
                    Orientation = Orientation.NORTH,
                    Coords = new Vec2i { X = 1, Y = 0 },
                });
            Simulation.Instance.Tick();

            Assert.AreEqual(testBuilding, visitor.Destination);
        }
    }
}
