using System.Collections.Generic;
using UnityEngine;
using UnityEngine.EventSystems;
using UnityEngine.UI;
using TMPro;

using Model.Cartography;
using Model.Common;
using Model.Inspection;
using Model.Landscape.Blueprints;

/// <summary>
/// Class responsible for placing buildings, and initializing the building placement menu based on config files.
/// </summary>
public class BuildingPlacer : SingletonMonobehaviour<BuildingPlacer>
{
    /// <summary>
    /// How many units in Unity should a cell in the model take up.
    /// </summary>
    [SerializeField] private float gridScale = 3f;
    public float GridScale { get => gridScale; }
    /// <summary>
    /// The prefab containing placeholders for building selection buttons.
    /// </summary>
    [SerializeField] private GameObject buildingMenuItemPrefab;

    /// <summary>
    /// Reference to the parent GameObject that contains the entertainment menu items.
    /// </summary>
    [SerializeField] private Transform entertainmentMenuContent;
    /// <summary>
    /// Reference to the parent GameObject that contains the food menu items.
    /// </summary>
    [SerializeField] private Transform foodMenuContent;
    /// <summary>
    /// Reference to the parent GameObject that contains the staff menu items.
    /// </summary>
    [SerializeField] private Transform staffMenuContent;
    /// <summary>
    /// Reference to the parent GameObject that contains the pavement menu items.
    /// </summary>
    [SerializeField] private Transform pavementMenuContent;

    /// <summary>
    /// Material used for displaying building ghosts when the selected building is allowed to be placed at the current position with the current rotation.
    /// </summary>
    [SerializeField] private Material placementAllowedMaterial;
    /// <summary>
    /// Material used for displaying building ghosts when the selected building is NOT allowed to be placed at the current position with the current rotation.
    /// </summary>
    [SerializeField] private Material placementRejectedMaterial;
    /// <summary>
    /// Raycast mask to prevent the ray used for finding the placement position from hitting buildings / ghosts.
    /// </summary>
    [SerializeField] private LayerMask placementRaycastMask;

    private IBuilder builder;

    private BuildingConfig selectedBuilding;
    private GameObject buildingGhost;
    private MeshRenderer ghostMeshRenderer; // the meshrenderer whose bounding box is used for finding the (0,0) corner.
    private List<Renderer> ghostRenderers; // Renderers to apply ghost materials on
    private int rotation = 0; // 0,1,2,3 for the 4 orientations

    [SerializeField] private string notEnoughMoneyMessage = "You don't have enough money to build this building!";

    /// <summary>
    /// Fill building selection menus and initialize builder.
    /// Note: We can't use the constructor, since that will be called by the engine
    /// and it's not guaranteed that all serialized references are set at that time.
    /// </summary>
    /// <param name="buildings">Array of building configs. These will be used to fill the building selection menu.</param>
    /// <param name="builder">Reference to the object used to communicate with the model about building placement. (Simulation)</param>
    public void Init(BuildingConfig[] buildings, IBuilder builder) {
        this.builder = builder;

        foreach (BuildingConfig bc in buildings) {
            GameObject g = Instantiate(buildingMenuItemPrefab);
            AssignParent(g.transform, bc.MenuCategory);

            Transform contents = g.transform.Find("Content");
            contents.Find("Image").GetComponent<Image>().sprite = bc.BuildingSelectorIcon;

            var bp = bc.Data.Blueprint as ObjectBlueprint;
            if (bp == null)
                continue;

            contents.Find("Name").GetComponent<TMP_Text>().text = bp.Name;
            contents.Find("Cost").GetComponent<TMP_Text>().text = bp.ConstructionCost.ToString().WithCurrency();


            // assign select command
            Button b = g.GetComponent<Button>();
            b.onClick.AddListener(() => {
                UIManager.Instance.ChangeMode(UIMode.BUILD);
                SelectBuilding(bc);
            });
        }
    }

    /// <summary>
    /// Select a building to place
    /// </summary>
    /// <param name="building">Config describing the building</param>
    private void SelectBuilding(BuildingConfig building) {
        var bp = building.Blueprint as ObjectBlueprint; // All building blueprints are ObjectBlueprints
        if (!CheckPrice(bp)) {
            MessageBox.Show(notEnoughMoneyMessage);
            return;
        }

        selectedBuilding = building;

        if (buildingGhost != null) // Destroy the ghost of the previously selected building
            Destroy(buildingGhost);

        // Create new building ghost
        buildingGhost = Instantiate(building.Prefab);
        buildingGhost.transform.rotation = Quaternion.Euler(0, rotation * -90.0f, 0);
        buildingGhost.transform.localScale = Vector3.one * gridScale;
        var db = buildingGhost.GetComponent<DisplayBehaviour>();
        db.enabled = false;
        ghostRenderers = db.Renderers;
        buildingGhost.GetComponent<Selectable>().enabled = false;
        ghostMeshRenderer = buildingGhost.GetComponentInChildren<MeshRenderer>();

        ghostRenderers.ForEach(x => x.sharedMaterial = placementRejectedMaterial);
    }

    /// <summary>
    /// Builds a building at a given position and orientation in the model and the view connecting the references between the two representations.
    /// </summary>
    /// <param name="building">The building to build</param>
    /// <param name="modelPlacement">Position and orientation of the building in the model</param>
    /// <param name="position">Position in unity</param>
    /// <param name="rotation">Rotation in unity</param>
    private void BuildBuilding(BuildingConfig building, MapPosition modelPlacement,
        Vector3 position, Quaternion rotation) {
        
        if (!CheckPrice(building.Blueprint as ObjectBlueprint)) {
            MessageBox.Show(notEnoughMoneyMessage);
            UIManager.Instance.ChangeMode(UIMode.NORMAL);
            return;
        }

        (IDisplayable displayable, IInspectable inspectable) = builder.Build(building.Blueprint, modelPlacement);

        GameObject g = Instantiate(building.Prefab, position, rotation);

        //Assign interfaces and initialize scripts on the spawned building
        DisplayBehaviour d = g.GetComponent<DisplayBehaviour>();
        Selectable s = g.GetComponent<Selectable>();

        d?.Init(displayable);
        s?.Init(inspectable);
    }

    private static bool CheckPrice(ObjectBlueprint bp) {
        return bp.ConstructionCost <= GameManager.Instance.PlayerBalance;
    }

    private void Update() {
        if (buildingGhost != null) {
            if (Input.mouseScrollDelta.y > 0.1f) {
                rotation++;
                buildingGhost.transform.rotation = Quaternion.Euler(0, rotation * -90.0f, 0);
            } else if (Input.mouseScrollDelta.y < -0.1f) {
                rotation--;
                buildingGhost.transform.rotation = Quaternion.Euler(0, rotation * -90.0f, 0);
            }

            Ray r = Camera.main.ScreenPointToRay(Input.mousePosition);
            RaycastHit hit;
            if (Physics.Raycast(r, out hit, float.PositiveInfinity, placementRaycastMask.value)) {
                DisplayGhostAndPlaceBuilding(hit.point);
            }
        }
    }

    /// <summary>
    /// Checks if the building can be placed at the given world position with the current orientation
    /// and places it if there was a mouseclick during the current frame.
    /// </summary>
    private void DisplayGhostAndPlaceBuilding(Vector3 worldPosition) {
        Vector2Int gridPos = WorldToGridPos(worldPosition) + selectedBuilding.GridOffset.Rotate90(rotation);
        Vector3 worldPos = GridToWorldPos(gridPos) + new Vector3(gridScale / 2, 0, gridScale / 2);
        buildingGhost.transform.position = worldPos;

        MapPosition modelPlacement = new MapPosition();
        modelPlacement.Coords = WorldToGridPos(ghostMeshRenderer.bounds.min + new Vector3(gridScale / 2, 0, gridScale / 2)).ToVec2i();
        modelPlacement.Orientation = rotation.ToOrientation();

        if (builder.CanBuild(selectedBuilding.Blueprint, modelPlacement)) {
            ghostRenderers.ForEach(x => x.sharedMaterial = placementAllowedMaterial);

            //EventSystem.current.IsPointerOverGameObject() returns true if pointer is over any gui element
            if (Input.GetMouseButtonUp(0) && !EventSystem.current.IsPointerOverGameObject()) {
                BuildBuilding(selectedBuilding, modelPlacement, worldPos, Quaternion.Euler(0, rotation * -90.0f, 0));
            }
        } else {
            ghostRenderers.ForEach(x => x.sharedMaterial = placementRejectedMaterial);
        }
    }

    private void OnEnable() {
        rotation = 0;
    }

    private void OnDisable() {
        if (buildingGhost != null)
            Destroy(buildingGhost);
        ghostMeshRenderer = null;
        ghostRenderers = null;
        buildingGhost = null;
        selectedBuilding = null;
    }

    private Vector2Int WorldToGridPos(Vector3 pos) {
        return new Vector2Int(Mathf.FloorToInt(pos.x / gridScale), Mathf.FloorToInt(pos.z / gridScale));
    }

    private Vector3 GridToWorldPos(Vector2Int pos) {
        return new Vector3(pos.x * gridScale, 0, pos.y * gridScale);
    }

    private void AssignParent(Transform t, BuildingCategory c) {
        switch (c) {
            case BuildingCategory.PAVEMENT:
                t.SetParent(pavementMenuContent, false);
                break;
            case BuildingCategory.ENTERTAINMENT:
                t.SetParent(entertainmentMenuContent, false);
                break;
            case BuildingCategory.FOOD:
                t.SetParent(foodMenuContent, false);
                break;
            case BuildingCategory.STAFF:
                t.SetParent(staffMenuContent, false);
                break;
            default:
                break;
        }
    }
}

/// <summary>
/// used to specify which menu to put a building in.
/// </summary>
public enum BuildingCategory
{
    PAVEMENT,
    ENTERTAINMENT,
    FOOD,
    STAFF,
}
