using System.Collections.Generic;
using UnityEngine;

using Model.Common;

/// <summary>
/// Class responsible for displaying a gameobject in the view for a given object in the model
/// </summary>
public class DisplayBehaviour : MonoBehaviour
{
    private IDisplayable displayedObject;
    [SerializeField] private List<Renderer> renderers;
    [SerializeField] private List<Collider> colliders;

    /// <summary>
    /// The minimal traveled squared distance needed to update the rotation.
    /// </summary>
    [Tooltip("The minimal traveled squared distance needed to update the rotation.")]
    [SerializeField] private float epsilon = 0.001f;

    /// <summary>
    /// Value for smoothing rotation. The lower the more smoothed (and delayed) the rotation is.
    /// Range: [0;1]
    /// </summary>
    [Range(0,1)]
    [SerializeField] private float rotationLerp = 0.1f;
    /// <summary>
    /// Value for smoothing position. The lower the more smoothed (and delayed) the rotation is.
    /// Range: [0;1]
    /// </summary>
    [Range(0, 1)]
    [SerializeField] private float positionLerp = 0.3f;

    /// <summary>
    /// The minimum squared distance traveled between two ticks to trigger a "teleport".
    /// A teleport is to move the object to the new position without smoothing.
    /// </summary>
    [SerializeField] private float teleportThresholdSquared = 0.25f;

    public List<Renderer> Renderers { get => renderers; }

    private bool visible;
    /// <summary>
    /// Note: setting this to a new value will update all renderers of the displaybehaviour.
    /// </summary>
    public bool Visible {
        get => visible;
        set {
            if (visible == value)
                return;
            visible = value;
            renderers.ForEach(x => x.enabled = value);
            colliders.ForEach(x => x.enabled = value);
        }
    }

    /// <summary>
    /// Initialization. Used instead of the constructor as that's called by the unity runtime therefore we don't have control over it.
    /// </summary>
    /// <param name="displayedObject">The interface responsible for communicating with the model.</param>
    public virtual void Init(IDisplayable displayedObject) {
        this.displayedObject = displayedObject;
        Visible = displayedObject.Visible;
        displayedObject.Handle.OnDeleted += (a,b) => Destroy(gameObject);
    }

    private void Update() {
        if (Visible = displayedObject.Visible) {
            Vector3 newPos = displayedObject.Position.ToVector3Scaled();
            float sqrDist = (transform.position - newPos).sqrMagnitude;

            if (sqrDist > epsilon) {
                if (sqrDist > teleportThresholdSquared)
                    transform.position = newPos;
                else
                    transform.position = Vector3.Lerp(transform.position, newPos, positionLerp*Time.deltaTime/GameManager.Instance.TickInterval);

                Quaternion lookRotation = Quaternion.LookRotation(newPos - transform.position, Vector3.up);
                transform.rotation = Quaternion.Slerp(transform.rotation,lookRotation,rotationLerp);
            }
        }
    }    
}
