using System.Collections;

using UnityEngine;

using Model.Common;
using Model.Inspection;

/// <summary>
/// Manager class responsible for:
///     - Setting up the simulation and ticking it
///     - Managing simulation speed
///     - Initializing other subsystems in the view. (ui manager, building placer)
///     - Spawning new people spawned by the model.
/// </summary>
public class GameManager : SingletonMonobehaviour<GameManager>
{
    [SerializeField] private float tickInterval = 1f / 30;
    public float TickInterval { get => tickInterval; }

    /// <summary>
    /// Target framerate is capped to prevent unnecessary calculations (and my laptop from being noisy) <br/>
    /// (For some reason VSync doesn't work despite being enabled.)
    /// </summary>
    [SerializeField] private int targetFrameRate = 60;

    [SerializeField] private UIManager uiManager;

    [SerializeField] private GameConfig gameConfig;
    /// <summary>
    /// Array of all buildable buildings, passed to buildingplacer to fill the building selection menu.
    /// </summary>
    [SerializeField] private BuildingConfig[] buildings;

    [SerializeField] private GameObject visitorPrefab;
    [SerializeField] private GameObject cleanerPrefab;
    [SerializeField] private GameObject maintainerPrefab;
    [SerializeField] private GameObject entrancePrefab;

    /// <summary>
    /// The Coroutine calling Tick() and running the simulation
    /// </summary>
    private IEnumerator tickRoutine;
    
    private long tickCount = 0;
    private bool isPaused = true;
    private Simulation simulation;

    /// <summary>
    /// Total ticks since the beginning of the simulation. Used for displaying a date.
    /// </summary>
    public long TickCount { get => tickCount; }
    public int PlayerBalance { get => simulation.GetMoney(); }
    public IInspectable Park { get => simulation.GetPark(); }


    public void SetSimulationSpeed(int ticksPerSec) {
        tickInterval = 1f / ticksPerSec;
    }

    public void MultiplySimulationSpeed(float times) {
        tickInterval /= times;
    }

    public void Pause() {
        if (isPaused)
            return;
        
        isPaused = true;
        StopCoroutine(tickRoutine);
    }

    public void Unpause() {
        if (!isPaused)
            return;

        isPaused = false;
        tickRoutine = Tick();
        StartCoroutine(tickRoutine);
    }

    public void Start() {
        Application.targetFrameRate = targetFrameRate;
        
        gameConfig.InsertConfigs();
        simulation = Simulation.CreateOnce(gameConfig.Config);
        uiManager.Init(buildings, simulation.GetBuilder());
        (IDisplayable d, IInspectable i) = simulation.GetEntrance();
        SpawnObject(entrancePrefab, d, i);
    }

    /// <summary>
    /// Tick the model and create the newly spawned people in the view.
    /// </summary>
    /// <returns></returns>
    public IEnumerator Tick() {
        while (!isPaused) {
            ++tickCount;
            TickResult res = simulation.Tick();
            if(res.newVisitors != null) // BUG: TickResult at the first tick is not initialized properly
            foreach ((IDisplayable d, IInspectable i) in res.newVisitors)
                SpawnObject(visitorPrefab, d, i);
            foreach ((IDisplayable d, IInspectable i) in res.newCleaningStaff)
                SpawnObject(cleanerPrefab, d, i);
            foreach ((IDisplayable d, IInspectable i) in res.newMaintenanceStaff)
                SpawnObject(maintainerPrefab, d, i);


            if(PlayerBalance < 0) {
                uiManager.DisplayGameOver();
                Pause();
            }

            yield return new WaitForSeconds(tickInterval);
        }
    }

    /// <summary>
    /// Spawn an object created by the model in the view.
    /// Currently used for people spawned by the model.
    /// </summary>
    /// <param name="prefab">The GameObject to instantiate in the view.</param>
    /// <param name="displayable">The displayable interface for keeping the position, rotation and visibility in sync.</param>
    /// <param name="inspectable">The inspectable interface for selectable objects.</param>
    private void SpawnObject(GameObject prefab, IDisplayable displayable, IInspectable inspectable = null) {
        GameObject g = Instantiate(prefab);

        DisplayBehaviour displayBehaviour = g.GetComponent<DisplayBehaviour>();
        if (displayBehaviour != null) {
            if (displayable != null) {
                displayBehaviour.Init(displayable);
            } else {
                Debug.LogError("Spawning displayable object with null as IDisplayable - Disabling DisplayBehaviour");
                displayBehaviour.enabled = false;
            }
        }

        Selectable selectable = g.GetComponent<Selectable>();
        if(selectable != null) {
            if (inspectable != null) {
                selectable.Init(inspectable);
            } else {
                Debug.LogError("Spawning selectable object with null as IInspectable - Disabling Selectable");
                selectable.enabled = false;
            }
        }
    }

    /// <summary>
    /// Destroys the simulation when the level is unloaded.
    /// This is necessary to prevent singletons in the model from leaking their state into the next game.
    /// </summary>
    private void OnDestroy() {
        Simulation.DestroyInstance();
    }
}
