using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;
using TMPro;

using Model.Utilities;
using Model.Inspection;
using System.Globalization;

/// <summary>
/// Class responsible for managing the inspector panel on the left
/// and displaying the properties of selected objects as well as providing the ability to edit them.
/// It also handles displaying and executing commands specific to the selected object.
/// </summary>
public class InspectorWindow : MonoBehaviour
{
    public bool IsOpen { get { return gameObject.activeSelf; } }

    /// <summary>
    /// Parent of the displayed fields and buttons.
    /// </summary>
    [SerializeField] private RectTransform contentRect;
    [SerializeField] private TMP_Text windowTitle;

    /// <summary>
    /// Template for displaying a non editable string field.
    /// In the GameObject tree of the prefab,
    /// the first TMP_Text component must be the one holding the name,
    /// the second must be the one holding the value.
    /// </summary>
    [SerializeField] private GameObject stringFieldPrefab;
    /// <summary>
    /// Template for displaying an editable string field.
    /// In the GameObject tree of the prefab,
    /// the first TMP_Text component must be the one holding the name,
    /// the second must be the one holding the value.
    /// </summary>
    [SerializeField] private GameObject editableStringFieldPrefab;
    /// <summary>
    /// Template for displaying a button for a command.
    /// In the GameObject tree of the prefab,
    /// the first TMP_Text component must be the one holding the name of the command.
    /// </summary>
    [SerializeField] private GameObject buttonPrefab;

    /// <summary>
    /// The color used for signaling invalid values in textfields.
    /// </summary>
    [SerializeField] private Color invalidValueBackgroundColor = new Color(0.51f, 0.114f, 0.188f); // #821D30

    /// <summary>
    /// Format provider for decimal separators.
    /// </summary>
    private static CultureInfo formatProvider;

    /// <summary>
    /// InspectorElements with registered eventhandlers that become invalid if the window is cleared.
    /// (Only Properties for now)
    /// </summary>
    private List<InspectionElement> registeredEvents = new List<InspectionElement>();

    /// <summary>
    /// Used to close the inspector if we select the same object again
    /// </summary>
    private IInspectable currentInspected;

    private void Awake() {
        formatProvider = CultureInfo.CurrentUICulture;
    }

    /// <summary>
    /// Creates and configures the UI elements for displaying the passed properties.
    /// </summary>
    /// <param name="properties"></param>
    private void ProcessProperties(IEnumerable<InspectionElement> properties) {
        if (properties == null)
            return;

        foreach(InspectionElement e in properties) {
            if (!e.IsVisible) {
                continue;
            }

            switch (e) {
                case Property<string> p:
                    CreateStringProperty(p);
                    break;
                case Property<int> p:
                    CreateGenericTextProperty(p, TMP_InputField.ContentType.IntegerNumber);
                    break;
                case Property<float> p:
                    CreateGenericTextProperty(p, TMP_InputField.ContentType.DecimalNumber, "0.00");
                    break;
                case Property<bool> p:
                    CreateBoolProperty(p);
                    break;
                case Property<RollingAggregation<int>> p:
                    CreateGenericToStringProperty(p);
                    break;
                default:
                    throw new NotImplementedException("Can't display this type of property");
            }
        }
    }

    /// <summary>
    /// Creates and configures the UI elements for displaying the passed commands.
    /// </summary>
    /// <param name="commands"></param>
    private void ProcessCommands(IEnumerable<Command> commands) {
        if (commands == null)
            return;

        foreach(Command c in commands)
            CreateCommandButton(c);
    }

    /// <summary>
    /// Open the window if it's closed and display the properties/commands of the passed IInspectable reference.
    /// </summary>
    /// <param name="inspectable"></param>
    public void Inspect(IInspectable inspectable) {
        if(currentInspected == inspectable) {
            Close();
            return;
        }

        windowTitle.text = inspectable.Name;

        ProcessProperties(inspectable.GetProperties());
        ProcessCommands(inspectable.GetCommands());

        currentInspected = inspectable;
        gameObject.SetActive(true);
    }

    /// <summary>
    /// Close the inspector window, clearing any created UI elements and deregistering their events.
    /// </summary>
    public void Close() {
        currentInspected = null;
        gameObject.SetActive(false);

        // Clear inspector
        for (int i = contentRect.childCount - 1; i >= 0; i--)
            Destroy(contentRect.GetChild(i).gameObject);

        foreach (var item in registeredEvents)
            item.ClearWeakEventHandlers();

        registeredEvents.Clear();
    }



    #region Methods for creating different types of property fields and buttons

    private void CreateGenericTextProperty<T>(Property<T> p,
        TMP_InputField.ContentType contentType = TMP_InputField.ContentType.Standard,
        string format = "") where T : IFormattable, IConvertible {

        GameObject g = Instantiate(p.IsEditable ? editableStringFieldPrefab : stringFieldPrefab);
        g.transform.SetParent(contentRect, false);
        var t = g.GetComponentsInChildren<TMP_Text>(); // t[0] is the name, t[1] is the value
        AssignText(t[0], p.Name);

        if (p.IsEditable) {
            var inputField = g.GetComponentInChildren<TMP_InputField>();
            var inputBackground = inputField.transform.GetComponent<Image>();
            Color originalBackgroundColor = inputBackground.color;

            AssignText(inputField, p.Value.ToString(format, formatProvider));
            p.RegisterWeakChangedEventHandler((_, val) => AssignText(inputField, val.ToString(format, formatProvider)));
            inputField.onEndEdit.AddListener((s) => {
                if (p.Edit((T)Convert.ChangeType(s, typeof(T))))
                    inputBackground.color = originalBackgroundColor;
                else
                    inputBackground.color = invalidValueBackgroundColor;
            });

            inputField.contentType = contentType;
        } else {
            AssignText(t[1], p.Value.ToString(format, formatProvider));
            p.RegisterWeakChangedEventHandler((_, val) => AssignText(t[1], val.ToString(format, formatProvider)));
        }
        registeredEvents.Add(p);
    }

    private void CreateStringProperty(Property<string> p) {
        GameObject g = Instantiate(p.IsEditable ? editableStringFieldPrefab : stringFieldPrefab);
        g.transform.SetParent(contentRect, false);
        var t = g.GetComponentsInChildren<TMP_Text>(); // t[0] is the name, t[1] is the value
        AssignText(t[0], p.Name);

        if (p.IsEditable) {
            var inputField = g.GetComponentInChildren<TMP_InputField>();
            var inputBackground = inputField.transform.GetComponent<Image>();
            Color originalBackgroundColor = inputBackground.color;

            AssignText(inputField, p.Value);
            p.RegisterWeakChangedEventHandler((_, val) => inputField.text = val);
            inputField.onEndEdit.AddListener((s) => {
                if (p.Edit(s))
                    inputBackground.color = originalBackgroundColor;
                else
                    inputBackground.color = invalidValueBackgroundColor;
            });
        } else {
            AssignText(t[1], p.Value);
            p.RegisterWeakChangedEventHandler((_, val) => AssignText(t[1], val));
        }
        registeredEvents.Add(p);
    }

    private void CreateGenericToStringProperty<T>(Property<T> p) {
        if (p.IsEditable)
            throw new NotSupportedException("Editing for generic toString properties is not supported");

        GameObject g = Instantiate(p.IsEditable ? editableStringFieldPrefab : stringFieldPrefab);
        g.transform.SetParent(contentRect, false);
        var t = g.GetComponentsInChildren<TMP_Text>(); // t[0] is the name, t[1] is the value
        AssignText(t[0], p.Name);

        AssignText(t[1], p.Value.ToString());
        p.RegisterWeakChangedEventHandler((_, val) => AssignText(t[1], val.ToString()));

        registeredEvents.Add(p);
    }

    private void CreateBoolProperty(Property<bool> p) {
        GameObject g = Instantiate(p.IsEditable ? editableStringFieldPrefab : stringFieldPrefab);
        g.transform.SetParent(contentRect, false);
        var t = g.GetComponentsInChildren<TMP_Text>(); // t[0] is the name, t[1] is the value
        AssignText(t[0], GetString(p.Name, p.Value));
        p.RegisterWeakChangedEventHandler((_, val) => AssignText(t[0], GetString(p.Name, val)));
        registeredEvents.Add(p);
        AssignText(t[1], "");

        if (p.IsEditable) {
            throw new NotImplementedException("Editable bool properties are not supported.");
        }

        static string GetString(string name, bool val) => val ? name : "Not " + name.ToLower();
    }

    private void CreateCommandButton(Command c) {
        GameObject g = Instantiate(buttonPrefab);
        g.transform.SetParent(contentRect, false);

        TMP_Text t = g.transform.GetComponentInChildren<TMP_Text>(); // t[0] is the name
        AssignText(t, c.Name);

        Button b = g.transform.GetComponentInChildren<Button>();
        if (!b)
            throw new NullReferenceException("Can't find Button on prefab");
        b.interactable = c.CanExecute;

        if (c.CommandSignature.Arguments.Count > 0) {
            b.onClick.AddListener(() => {
                UIManager.Instance.ExecuteWithParameter(c);
            });
        } else {
            b.onClick.AddListener(
            () => {
                if (c.CanExecute)
                    c.Execute();

                b.interactable = c.CanExecute;
            });
        }
    }

    private void AssignText(TMP_Text t, string text) {
        if (t)
            t.text = text;
        else
            throw new NullReferenceException("Can't find text component");
    }

    private void AssignText(TMP_InputField t, string text) {
        if (t)
            t.text = text;
        else
            throw new NullReferenceException("Can't find text component");
    }

    #endregion
}
