using System;
using System.Collections.Generic;
using UnityEngine;
using UnityRandom = UnityEngine.Random; // Avoid collision with System.Random

using Model.Common;
using Model.Landscape;
using Model.Inspection;

/// <summary>
/// Class responsible for pavement building specific things in the view.
/// </summary>
public class PavementBehaviour : DisplayBehaviour
{
    /// <summary>
    /// The garbage bin prefab to instantiate when the player buys a bin.
    /// </summary>
    [Header("Pavement specific")]
    [SerializeField] private GameObject bin;
    /// <summary>
    /// List of garbage props to chose randomly from.
    /// </summary>
    [SerializeField] private List<GameObject> garbagePrefabs;

    /// <summary>
    /// Bottom left (closer to 0,0,0) corner of the rectangle in which garbage is displayed.
    /// </summary>
    [SerializeField] private Vector3 garbageAreaMin;
    /// <summary>
    /// Top right (closer to 1,1,1) corner of the rectangle in which garbage is displayed.
    /// </summary>
    [SerializeField] private Vector3 garbageAreaMax;

    private Pavement pavement;
    public Pavement Pavement { get => pavement; } // used for cleaner destination selection

    private Queue<GameObject> spawnedGarbage = new Queue<GameObject>();

    /// <summary>
    /// Initialize base class and register event handlers.
    /// </summary>
    /// <param name="displayedObject"></param>
    public override void Init(IDisplayable displayedObject) {
        base.Init(displayedObject);

        pavement = displayedObject as Pavement;
        if(pavement == null)
            throw new ArgumentException("PavementBehaviour must be initialized with a Pavement reference");

        pavement.HasGarbageCan.Changed += DisplayBin;
        pavement.GarbageAmount.Changed += ChangeDisplayedGarbage;
    }

    /// <summary>
    /// Deregister events whenever a pavement is destroyed.
    /// Not used currently since buildings can't be removed.
    /// </summary>
    private void OnDestroy() {
        if (pavement != null) { // necessary because of building ghost during placement
            pavement.HasGarbageCan.Changed -= DisplayBin;
            pavement.GarbageAmount.Changed -= ChangeDisplayedGarbage;
        }
    }

    private void DisplayBin(Property<bool> sender, bool value) {
        bin.SetActive(value);
    }

    /// <summary>
    /// Update the spawned trash props creating new or destroying old ones to match the amount of garbage in the model.
    /// </summary>
    /// <param name="source">unused, required by the signature of the event</param>
    /// <param name="newGarbageAmount"></param>
    private void ChangeDisplayedGarbage(Property<int> source, int newGarbageAmount) {
        while (spawnedGarbage.Count > 0 && newGarbageAmount < spawnedGarbage.Count)
            Destroy(spawnedGarbage.Dequeue());

        while (spawnedGarbage.Count < newGarbageAmount) {
            Vector3 randPos = new Vector3(
                UnityRandom.Range(garbageAreaMin.x, garbageAreaMax.x),
                UnityRandom.Range(garbageAreaMin.y, garbageAreaMax.y),
                UnityRandom.Range(garbageAreaMin.z, garbageAreaMax.z));

            Quaternion randRot = Quaternion.Euler(0f, UnityRandom.Range(0f, 360f), 0f);

            GameObject go = Instantiate(garbagePrefabs[UnityRandom.Range(0, garbagePrefabs.Count)]);
            go.transform.SetParent(transform);

            go.transform.localPosition = randPos;
            go.transform.localRotation = randRot;

            spawnedGarbage.Enqueue(go);
        }
    }

    /// <summary>
    /// Draw the garbage area for debugging in the editor. (Only draws if the pavement is selected in the editor.)
    /// </summary>
    private void OnDrawGizmosSelected() {
        Gizmos.color = Color.red;
        Gizmos.DrawWireCube(transform.position + (garbageAreaMax + garbageAreaMin) / 2, garbageAreaMax - garbageAreaMin);
    }
}
