using System.Collections.Generic;
using UnityEngine;

using Model.Inspection;

/// <summary>
/// Provides an example and serves as a testbed for the inspection system.
/// </summary>
public class ExampleInspectable : MonoBehaviour, IInspectable
{
    /// <summary>
    /// Some test values for basic string properties generated in Start().
    /// </summary>
    [SerializeField]
    private List<string> testValues = new List<string>();

    /// <summary>
    /// Keep the string values in the editor in sync with the properties while the game is running.
    /// </summary>
    [SerializeField]
    [Tooltip("Keep the string values in the editor in sync with the properties")]
    private bool updateIfChanged = true;

    /// <summary>
    /// Title of the inspector window when this object is selected.
    /// </summary>
    public string Name => "Example Inspectable";

    private List<InspectionElement> properties = new List<InspectionElement>();
    private List<Command> commands = new List<Command>();

    private Property<int> p0 = new Property<int>("meaning of life", 42, true);
    private Property<string> p1 = new Property<string>("EDITABLE STRING", "text", true);
    private Property<int> p2 = new Property<int>("Readonly INT", 42);
    private Property<int> p3 = new Property<int>("Editable INT", 420, true);
    private Property<float> p4 = new Property<float>("PI", 3.14159f);
    private Property<float> p5 = new Property<float>("Editable FLOAT", 1.2345f, true);

    public IEnumerable<InspectionElement> GetProperties() {
        var l = new List<InspectionElement>(properties);
        l.AddRange(Inspector.GetProperties(this));
        return l;
    }
    public IEnumerable<Command> GetCommands() => commands;

    private void Start() {
        GetComponent<Selectable>().Init(this);

        foreach (var s in testValues) {
            properties.Add(new Property<string>(s.ToUpper(), s));
        }

        commands.Add(new Command(
            "Hello World",
            (x) => { Debug.Log("Hello World!"); return null; },
            Command.Signature.Action,
            () => true));
        bool once = true;
        commands.Add(new Command(
            "Once",
            (x) => { Debug.Log("Once"); once = false; return null; },
            Command.Signature.Action,
            () => once));
        commands.Add(new Command(
            "MessageBox",
            (x) => { MessageBox.Show("Hello MessageBox!"); return null; },
            Command.Signature.Action,
            () => true ));
    }

    private void Update()
    {
        if (updateIfChanged) { // Apply changes from the editor's inspector (at runtime) to the properties.
            for (int i = 0; i < testValues.Count; i++) {
                var p = properties[i] as Property<string>;
                if (testValues[i] != p.Value)
                    p.Value = testValues[i];
            }
        }
    }
}
