using System.Collections;
using System.Collections.Generic;
using UnityEngine;

using Model.Inspection;

/// <summary>
/// Class responsible for selecting objects in the view.
/// </summary>
public class Selectable : MonoBehaviour
{
    [SerializeField] private Material selectedMaterial;

    [SerializeField] private List<Renderer> selectionRenderers;

    private List<Material> baseMaterials;
    private IInspectable inspectable;

    /// <summary>
    /// Saves materials of all renderers so that they can be restored when the object is deselected.
    /// </summary>
    private void Awake() {
        baseMaterials = new List<Material>(selectionRenderers.Count);
        selectionRenderers.ForEach((x) => baseMaterials.Add(x.sharedMaterial));
    }

    /// <summary>
    /// Set the IInspectable used for displaying properties when selected.
    /// </summary>
    /// <param name="inspectable"></param>
    public void Init(IInspectable inspectable) {
        this.inspectable = inspectable;
    }

    /// <summary>
    /// Selects the object and updates materials accordingly.
    /// </summary>
    public void Select() {
        selectionRenderers.ForEach((x) => x.sharedMaterial = selectedMaterial);
        UIManager.Inspector.Inspect(inspectable);
    }

    /// <summary>
    /// Change materials back to normal and close the inspector
    /// </summary>
    public void Deselect() {
        var renderers = selectionRenderers.GetEnumerator();
        var materials = baseMaterials.GetEnumerator();
        while (renderers.MoveNext() && materials.MoveNext()) {
            renderers.Current.sharedMaterial = materials.Current;
        }
        UIManager.Inspector.Close();
    }
}
