using System;
using System.Linq;
using UnityEngine;
using UnityEngine.EventSystems;
using TMPro;

using Model.Common;
using Model.Inspection;
using Model.Landscape;

/// <summary>
/// Class responsible for displaying miscellaneous information as well as orchestrating the different UI modes.
/// </summary>
public class UIManager : SingletonMonobehaviour<UIManager>
{
    [SerializeField] private TMP_Text dateText;
    /// <summary>
    /// The text component for displaying the player's money.
    /// </summary>
    [SerializeField] private TMP_Text balanceText;
    [SerializeField] private int ticksPerDay;
    [SerializeField] private BuildingPlacer buildingPlacer;

    /// <summary>
    /// in-game time displayed at the bottom right corner
    /// </summary>
    private DateTime date;
    private UIMode mode;
    private Selectable currentSelected;

    [SerializeField] private InspectorWindow inspector;
    public static InspectorWindow Inspector { get => Instance.inspector; }


    // stuff for SELECT_PARAMETER mode
    private Command commandRequestingParameters;
    /// <summary>
    /// The gameobject that indicates that the UI is in parameter selection mode.
    /// Currently a "Cancel parameter selection" button.
    /// </summary>
    [SerializeField] private GameObject selectModeNotifier;

    // Game Over related
    [SerializeField]
    private GameObject gameOverParent;

    /// <summary>
    /// Initialize the UI
    /// </summary>
    /// <param name="buildings"></param>
    /// <param name="builder"></param>
    public void Init(BuildingConfig[] buildings, IBuilder builder) {
        buildingPlacer.Init(buildings, builder);
    }

    /// <summary>
    /// Changes the UI mode, enables or disables the building placer and the selectModeNotifier according to the new UI mode.
    /// </summary>
    /// <param name="newMode"></param>
    public void ChangeMode(UIMode newMode) {
        if (mode == newMode)
            return;

        // Note: enabling and disabling the buildingPlacer has side effects so it should be prevented if not necessary.
        if (mode == UIMode.BUILD)
            buildingPlacer.enabled = false;
        else if (newMode == UIMode.BUILD)
            buildingPlacer.enabled = true;

        selectModeNotifier.SetActive(newMode == UIMode.SELECT_PARAMETER);

        mode = newMode;
    }

    /// <summary>
    /// Instructs the UIManager to enter parameter selection mode for the given command and execute it once the parameters are collected.
    /// </summary>
    /// <param name="command"></param>
    public void ExecuteWithParameter(Command command) {
        this.commandRequestingParameters = command;
        ChangeMode(UIMode.SELECT_PARAMETER);
    }

    /// <summary>
    /// Display the details of the park (this is the method called when the player clicks on the money in the lower left)
    /// </summary>
    public void InspectPark() {
        if(currentSelected != null)
            ClearSelection();
        inspector.Inspect(GameManager.Instance.Park);
    }

    public void DisplayGameOver() {
        gameOverParent.SetActive(true);
    }

    private void Start() {
        date = DateTime.Now;
    }

    private void Update() {
        // update simple text fields
        dateText.text = date.AddDays((double)GameManager.Instance.TickCount / ticksPerDay).ToString("yyyy.MM.dd");
        balanceText.text = GameManager.Instance.PlayerBalance.ToString().WithCurrency();

        // !EventSystem.current.IsPointerOverGameObject() prevents selecting objects behind UI elements
        if (Input.GetMouseButtonDown(0) && !EventSystem.current.IsPointerOverGameObject()) {
            Ray ray = Camera.main.ScreenPointToRay(Input.mousePosition);
            RaycastHit info;
            if (Physics.Raycast(ray, out info)) {
                if (mode == UIMode.NORMAL) {
                    Selectable s = info.collider.gameObject.GetComponent<Selectable>();
                    Select(s);
                } else if(mode == UIMode.SELECT_PARAMETER) {
                    if (TryCommandWithParameter(info.collider.gameObject))
                        ChangeMode(UIMode.NORMAL);
                    else
                        MessageBox.Show("Invalid parameter");
                }
            } else {
                if (mode == UIMode.NORMAL)
                    ClearSelection();
                else if (mode == UIMode.SELECT_PARAMETER)
                    MessageBox.Show("Invalid parameter");
            }
        }

        // Esc = deselect (and close inspector)
        if (Input.GetKeyDown(KeyCode.Escape)) {
            ClearSelection();
        }
    }

    private void Select(Selectable s) {
        if (currentSelected == s || s == null) { // Deselect if we select the same object again
            ClearSelection();
            return;
        }

        if (currentSelected)
            currentSelected?.Deselect();
        else if (inspector.IsOpen)
            inspector.Close();

        currentSelected = s;
        currentSelected.Select();
    }

    private void ClearSelection() {
        if (currentSelected)
            currentSelected.Deselect(); // also closes inspector
        else
            inspector.Close(); // if park is currently inspected
        currentSelected = null;
    }

    /// <summary>
    /// Check if the passed GameObject can be the argument of the currently pending command.
    /// If yes, execute the command with it and return to NORMAL UI mode, do nothing otherwise.
    /// </summary>
    /// <param name="target"></param>
    /// <returns>Whether the passed GameObject was compatible with the pending command.
    /// If yes, the command has been executed and the UI is switched back to NORMAL mode.</returns>
    private bool TryCommandWithParameter(GameObject target) {
        if (commandRequestingParameters.CommandSignature.Arguments.Count > 1)
            throw new NotImplementedException("Multiple arguments are not supported");

        Type argType = commandRequestingParameters.CommandSignature.Arguments.Single();
        switch (argType) {
            case var p when p == typeof(Pavement): //workaround for using switch with typeof
                PavementBehaviour pb = target.GetComponent<PavementBehaviour>();
                if (pb == null)
                    return false;

                if (commandRequestingParameters.CanExecute) {
                    commandRequestingParameters.Execute(pb.Pavement);
                    return true;
                } else
                    return false;

            default:
                return false;
        }
    }
}

public enum UIMode
{
    NORMAL, // Select & inspect entities.
    BUILD,  // For placing buldings
    SELECT_PARAMETER, // Select parameter for a pending command.
}
